#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Globals file for dms system
"""

from magcode.core.globals_ import settings

# settings for where files are
settings['config_dir'] = '/etc/dms'
settings['log_dir'] = '/var/log/dms'
settings['run_dir'] = '/var/run/dms'
settings['var_lib_dir'] = '/var/lib/dms'
settings['config_file'] = settings['config_dir'] + '/' + 'dms.conf'
# DMS only uses one daemon 
settings['pid_file'] = settings['run_dir'] + '/' + 'dmsdmd.pid'
settings['event_queue_pid_file'] = settings['pid_file']
settings['log_file'] = settings['log_dir'] \
        + '/' + settings['process_name'] + '.log'
settings['panic_log'] = settings['log_dir'] \
        + '/' + settings['process_name'] + '-panic.log'

# settings initialisations that cause trouble with config key checking
# if in /etc/dms.conf [DEFAULT] section

# DB settings to help prevent RAM piggery
settings['db_query_slice'] = 1000
settings['preconvert_int_settings'] += 'db_query_slice'
# DB event queue columns
# DO NOT CHANGE THIS UNLESS YOU KNOW WHAT YOU ARE DOING!
settings['event_queue_fkey_columns'] = 'zone_id server_id master_id'

# dmsdmd.py
# Print debug mark
settings['debug_mark'] = False
# Number of seconds we wait while looping in main loop...
settings['sleep_time'] = 3 # seconds
settings['debug_sleep_time'] = 20 # seconds
settings['memory_exec_threshold'] = 250 #MB

# dyndns_update.py
settings['dig_path'] = 'dig'
settings['dig_arguments'] = "+noall +answer"
settings['rndc_path'] = "rndc"
settings['rndc_arguments'] = ""
settings['nsupdate_path'] = "rndc"
settings['nsupdate_arguments'] = ""
# The following works because of the import at the top!
if (settings['os_family'] == 'FreeBSD'):
    settings['dyndns_key_file'] = '/etc/namedb/update-session.key'
elif (settings['os_family'] == 'Linux'):
    settings['dyndns_key_file'] = '/etc/dms/bind/update-session.key'
else:
    settings['dyndns_key_file'] = '/etc/bind/update-session.key'
settings['dyndns_key_name'] = 'update-ddns'
settings['dnssec_filter'] = 'RRSIG DNSKEY NSEC3 NSEC3PARAM TYPE65534 NSEC'
settings['dyndns_success_rcodes'] = 'NOERROR'
settings['dyndns_retry_rcodes'] = 'SERVFAIL NOTAUTH NXRRSET'
settings['dyndns_reset_rcodes'] = 'NXDOMAIN'
settings['dyndns_fatal_rcodes'] = 'BADVERS NOTIMP NOTZONE YXRRSET YXDOMAIN FORMERR REFUSED'

# update_engine.py
settings['dns_server'] = 'localhost'
settings['dns_port'] = 'domain'
settings['dns_query_timeout'] = 30 # seconds
settings['soa_serial_wrap_threshold'] = 9950
settings['nsec3_salt_bit_length'] = 64
settings['nsec3_hash_algorithm'] = 1
settings['nsec3_flags'] = 1
settings['nsec3_iterations'] = 10
# Place for Update engines fo be registered.
update_engine = {}

# zone_text_utils.py
settings['apex_rr_tag'] = 'APEX_RRS'
settings['comment_group_leader'] = ';|'
settings['comment_rr_leader'] = ';#'
settings['comment_rrflags_leader'] = ';!'
# Don't use '^' at the start of the following regexp - comment line will cause
# the zone_parser to fail!
settings['comment_anti_regexp'] = ';[^!#|]'
settings['rr_flag_lockptr'] = 'LOCKPTR'
settings['rr_flag_forcerev'] = 'FORCEREV'
settings['rr_flag_disable'] = 'DISABLE'
settings['rr_flag_ref'] = 'REF:'
settings['rr_flag_rrop'] = 'RROP:'
settings['rr_flag_trackrev'] = 'TRACKREV'

#zone_cfg.py
settings['apex_ns_key'] = 'apex_ns'

# zone_sm.py
settings['apex_comment_template'] = 'Apex resource records for %s'
settings['edit_lock_timeout'] = 30 # minutes
if (settings['os_family'] == 'Linux'):
    settings['dms_bind_config_dir'] = settings['config_dir'] + '/' + 'bind'
    settings['master_bind_config_dir'] = '/etc/bind'
    settings['master_config_dir'] = settings['var_lib_dir'] \
                                     + '/' + 'master-config'
    settings['master_dyndns_dir'] = '/var/lib/bind/dynamic'
    settings['master_slave_dir'] = '/var/lib/bind/slave'
    settings['master_static_dir'] = '/var/lib/bind/static'
    settings['master_dnssec_key_dir'] = '/var/lib/bind/keys'
elif (settings['os_family'] == 'FreeBSD'):
    settings['master_bind_config_dir'] = '/etc/namedb'
    settings['dms_bind_config_dir'] = settings['master_bind_config_dir']
    settings['master_config_dir'] = '/etc/namedb/master-config'
    settings['master_dyndns_dir'] = '/etc/namedb/dynamic'
    settings['master_slave_dir'] = '/etc/namedb/slave'
    settings['master_static_dir'] = '/etc/namedb/static'
    settings['master_dnssec_key_dir'] = '/etc/namedb/keys'
settings['master_template_dir_name'] = 'master-config-templates'
settings['master_template_dir'] = settings['config_dir'] + '/' \
                                        + settings['master_template_dir_name']
MASTER_DYNDNS_TEMPLATE = 'master_dyndns_template'
settings[MASTER_DYNDNS_TEMPLATE] = 'dynamic-config.conf'
MASTER_AUTO_DNSSEC_TEMPLATE = 'master_auto_dnssec_template'
settings[MASTER_AUTO_DNSSEC_TEMPLATE] = 'auto-dnssec-config.conf'
MASTER_SERVER_ACL_TEMPLATE = 'master_server_acl_template'
settings[MASTER_SERVER_ACL_TEMPLATE] = 'server-acl.conf'
MASTER_STATIC_TEMPLATE = 'master_static_template'
settings[MASTER_STATIC_TEMPLATE] = 'static-config.conf'
MASTER_SLAVE_TEMPLATE = 'master_slave_template'
settings[MASTER_SLAVE_TEMPLATE] = 'slave-config.conf'
# This is a file name as a tempfile is created and mved into place
settings['master_include_file'] = settings['master_config_dir'] \
                                    + '/' + 'zones.conf' 
settings['master_server_acl_file'] = settings['master_config_dir'] \
                                    + '/' + 'server-acl.conf' 
settings['acl_name_extension'] = '-servers'
settings['default_acl_name'] = 'dms' + settings['acl_name_extension']
settings['zone_file_mode'] = '00664'
settings['zone_file_group'] = 'bind'

# master_sm.py
settings['master_hold_timeout'] = 10  #minutes
settings['master_rndc_settle_delay'] = 5 #seconds
settings['config_file_mode'] = '00644'

# server_group.py
settings['server_config_dir'] = settings['config_dir'] \
                + '/server-config-templates'
settings['server_replica_suffix'] = '-replica'
settings['sg_config_dir'] = settings['var_lib_dir'] + '/dms-sg'
# These 2 settings are initialized in dms/apps/dmsdmd.py
settings['master_dns_server'] = None
settings['master_dns_port'] = 'domain'
settings['connect_retry_wait'] = 10
settings['this_servers_addresses'] = []

# server_sm.py
settings['rsync_path'] = 'rsync'
settings['rsync_password_file'] = settings['config_dir'] \
                + '/' + 'rsync-dnsconf-password'
settings['rsync_args'] = '--quiet -rptv'
settings['rsync_target_user'] = 'dnsconf'
settings['rsync_target_module'] = 'dnsconf'
settings['rsync_target'] = (settings['rsync_target_user'] + '@%s::' 
                + settings['rsync_target_module'] + '/')
settings['rsync_dnssec_args'] = '--quiet -av'
settings['rsync_dnssec_password_file'] = settings['config_dir'] \
                                + '/' + 'rsync-dnssec-password'
settings['rsync_dnssec_user'] = 'dnssec'
settings['rsync_dnssec_module'] = 'dnssec'
settings['rsync_dnssec_target'] = (settings['rsync_dnssec_user'] + '@%s::' 
                            + settings['rsync_dnssec_module'] + '/')
settings['serversm_soaquery_success_rcodes'] = 'NOERROR'
settings['serversm_soaquery_ok_rcodes'] = settings['serversm_soaquery_success_rcodes'] + ' ' + 'NXDOMAIN REFUSED NOTAUTH'
settings['serversm_soaquery_retry_rcodes'] = ''
settings['serversm_soaquery_broken_rcodes'] = 'SERVFAIL FORMERR BADVERS NOTIMP YXDOMAIN YXRRSET NXRRSET'
settings['serversm_soaquery_domain'] = 'localhost.'
settings['bind9_zone_count_tag'] = 'number of zones:'

# Administrative security role tag
settings['admin_sectag'] = 'Admin'
# Accept GET Method for JSON RPC request(s)
# Some old JSON RPC libraries might need this...
settings['jsonrpc_accept_get'] = False

# cmdline_engine
settings['zone_del_off_age'] = 1000 * 366 # days
settings['preconvert_float_settings'] += 'zone_del_off_age'

# zone_engine.py
settings['list_events_last_limit'] = 25

# zone_tool.py
# admin_group_list shifted to magcode.core.globals_
settings['restricted_mode_commands'] = 'clear_edit_lock create_zone copy_zone copy_zi delete_zone diff_zone_zi diff_zones disable_zone enable_zone edit_zone exit EOF help ls ls_deleted ls_reference ls_zi quit refresh_zone refresh_zone_ttl reset_zonesm record_query_db show_apex_ns show_config show_dms_status show_zi show_zone show_zone_byid show_zone_sectags show_zonesm show_zonesm_byid undelete_zone'
settings['wsgi_test_commands'] = 'create_zi_zone cancel_edit_zone rr_query_db set_zone_alt_sg show_zi_byid list_zone list_zone_deleted list_zi'
# Command log facility
settings['commands_not_to_syslog'] = 'help show EOF list quit exit rr ls'
settings['zone_tool_log_facility'] = 'local7'
settings['zone_tool_log_level'] = 'info'
# zone_tool rndc  and key file stuff
settings['config_template_dir'] = settings['config_dir'] + '/config-templates'
settings['rndc_header_template'] = settings['config_template_dir'] \
                                    + '/rndc.conf-header'
settings['rndc_server_template'] = settings['config_template_dir'] \
                                    + '/rndc.conf-server'
settings['rndc_conf_file'] = settings['var_lib_dir'] + '/rndc' + '/rndc.conf'
settings['rndc_conf_file_mode'] = '00644'
settings['server_admin_config_dir'] = settings['config_dir'] \
                                        + '/server-admin-config'
settings['tsig_key_template'] = settings['config_template_dir'] \
                                    + '/tsig.key'
settings['key_file_mode'] = '00640'
settings['key_file_owner'] = 'root'
settings['key_file_group'] = 'bind'
# zone_tool ping stuff
settings['oping_path'] = 'oping'
settings['oping_args'] = '-i 0.2 -c 5'

# auto reverse control settings
settings['auto_reverse'] = True
settings['auto_create_ipv4_ptr'] = False
settings['auto_create_ipv6_ptr'] = True
settings['preconvert_bool_settings'] += 'auto_reverse auto_create_ipv4_ptr auto_create_ipv6_ptr'




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
DNS Zone State Machine
"""


import glob
import os.path
from tempfile import mkstemp
import io
import os
import errno
import grp
import pwd
from datetime import timedelta
from datetime import datetime

from sqlalchemy.orm import reconstructor
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql import and_
from sqlalchemy.sql import or_
from sqlalchemy.sql import not_

from magcode.core.database import *
from dms.exceptions import *
from dms.dns import RRCLASS_IN
from dms.globals_ import update_engine
from dms.globals_ import MASTER_STATIC_TEMPLATE
from dms.globals_ import MASTER_SLAVE_TEMPLATE
from dms.globals_ import MASTER_DYNDNS_TEMPLATE
from dms.globals_ import MASTER_AUTO_DNSSEC_TEMPLATE
import dms.database.zone_instance
import dms.database.zone_sectag
import dms.database.server_group
import dms.database.update_group
import dms.database.reverse_network
from magcode.core.database.event import create_event
from magcode.core.database.event import cancel_event
from magcode.core.database.event import reschedule_event
from magcode.core.database.state_machine import StateMachine
from magcode.core.database.state_machine import SMEvent
from magcode.core.database.state_machine import SMSyncEvent
from magcode.core.database.state_machine import StateMachineError
from magcode.core.database.state_machine import StateMachineFatalError
from magcode.core.database.state_machine import smregister
from magcode.core.database.event import Event
from magcode.core.database.event import ESTATE_SUCCESS
from magcode.core.database.event import eventregister
from magcode.core.database.event import synceventregister
from magcode.core.database.event import queue_event
from dms.template_cache import read_template
import dms.database.zone_cfg as zone_cfg
from dms.zone_text_util import data_to_bind
from dms.database.master_sm import zone_sm_reconfig_schedule
from dms.database.master_sm import zone_sm_dnssec_schedule
from dms.database.master_sm import reconfig_sg
from dms.database.sg_utility import find_sg_byname
from dms.database.sg_utility import list_all_sgs
from dms.database.reference import new_reference
from dms.exceptions import NoSgFound


sql_data['zone_sm_subclasses'] = []
sql_data['zone_sm_type_list'] = []

# Some Static Constants
ZSTATE_INIT = 'INIT'
ZSTATE_CREATE = 'CREATE'
ZSTATE_RESET = 'RESET'
ZSTATE_BATCH_CREATE_1 = 'BATCH_CREATE_1'
ZSTATE_BATCH_CREATE_2 = 'BATCH_CREATE_2'
ZSTATE_DELETED = 'DELETED'
ZSTATE_UNCONFIG = 'UNCONFIG'
ZSTATE_DISABLED = 'DISABLED'
ZSTATE_UPDATE = 'UPDATE'
ZSTATE_PUBLISHED = 'PUBLISHED'
ZLSTATE_EDIT_UNLOCK = 'EDIT_UNLOCK'
ZLSTATE_EDIT_LOCK = 'EDIT_LOCK'
zone_sm_states = (ZSTATE_UNCONFIG, ZSTATE_DISABLED,
                    ZSTATE_PUBLISHED, ZSTATE_UPDATE)

# Zone State machine exceptions
# Error Exceptions
class ZoneSMError(StateMachineError):
    """
    Base exception for Zone State Machine.
    """
    pass

class ZoneSMFatalError(StateMachineFatalError):
    """
    Base exception for Zone State Machine Fatal errors.
    """
    pass
class ZoneSMLockFailure(ZoneSMFatalError):
    """
    Lock mechanism failure
    """
class ZoneSMEventFailure(ZoneSMFatalError):
    """
    Event failure
    """
class ZoneSMUpdateFailure(ZoneSMFatalError):
    """
    Update failure
    """
class ZoneEditLocked(ZoneSMFatalError):
    """
    Exception raised if zone Locked, and an attempt
    to update or Exit the Edit session is made.
    """
    def __init__(self, domain, edit_lock_token, locked_by, locked_at):
        if locked_by:
            message = (
                "Zone '%s' is locked with token '%s', held by '%s', since %s."
                        % (domain, edit_lock_token, locked_by, locked_at))
        else:
            message = ("Zone '%s' is locked with token '%s' since %s."
                        % (domain, edit_lock_token, locked_at))
        Exception.__init__(self, message)
        self.domain = domain
        self.edit_lock_token = edit_lock_token

class ZoneEditLockTimedOut(ZoneSMFatalError):
    """
    Exception raised if zone was locked, and then timed out
    and an attempt to update zone ior tickle the edit lock is made.
    """
    def __init__(self, domain):
        message = ("Zone '%s' was locked and the lock timed out."
                        % domain )
        Exception.__init__(self, message)
        self.domain = domain

class ZoneNotEditLocked(ZoneSMFatalError):
    """
    Exception raised if zone Locked, and an attempt
    to update or Exit the Edit session is made.
    """
    def __init__(self, domain):
        message = ("Zone '%s' is not LOCKED."
                    % (domain))
        Exception.__init__(self, message)
        self.domain = domain

class ReadSoaRetry(ZoneSMError):
    """
    read_soa failed because of a temporary error 
    Will retry.
    """
class ZoneInBindRetry(ZoneSMError):
    """
    Retrying file configuration until bind reconfigured without zone.
    """
class ReadSoaFailed(ZoneSMFatalError):
    """
    read_soa failed permanently.
    """
class DynDNSUpdateRetry(ZoneSMError):
    """
    Dynamic Zone update failed because of a temporary error 
    Will retry.
    """
class DynDNSUpdateFailed(ZoneSMFatalError):
    """
    Dynamic Zone update failed permanently.
    """
class BatchCreateFailed(ZoneSMFatalError):
    """
    Dynamic Zone update failed permanently.
    """
class CreateFailed(ZoneSMFatalError):
    """
    Dynamic Zone update failed permanently.
    """
class ZoneAlreadyDisabled(ZoneSMFatalError):
    """
    Disabled zone is already disabled.
    """
class ZoneAlreadyEnabled(ZoneSMFatalError):
    """
    Disabled zone is already disabled.
    """
class ZoneResetDisabled(ZoneSMFatalError):
    """
    Disabled zone can't be reset
    """
class ZoneSMUndeleteFailure(ZoneSMFatalError):
    """
    Undelete failure - No ZI or Active zone exists
    """
class ZoneNotDestroyedFilesExist(ZoneSMFatalError):
    """
    Can't destroy a Zone as its Bind files still exist
    """
class ZoneNoAltSg(ZoneSMFatalError):
    """
    Can't swap SGs as alt_sg is not set.
    """

# Zone State Machine events
@eventregister
class ZoneSMBatchConfig(SMEvent):
    """
    Initialize configuration during a batch load
    """
@eventregister
class ZoneSMReconfigCheck(SMEvent):
    """
    Check that Master NS has loaded zone
    """
@eventregister
class ZoneSMConfig(SMEvent):
    """
    For ordinary configuration of zones.  Writes out initial ZI to 
    master dynamic directory, and then issues ZoneSmReconfigUpdate
    """
@eventregister
class ZoneSMReconfigUpdate(SMEvent):
    """
    Refresh zone on Master NS reconfig, loadkeys, or signzone
    """

@eventregister
class ZoneSMUpdate(SMEvent):
    """
    Update event publishing a ZI
    """
@eventregister
class ZoneSMRemoveZoneFiles(SMEvent):
    """
    Delete Zone files for a disabled or deleted domain
    """
@synceventregister
class ZoneSMDoReset(SMSyncEvent):
    """
    Reset the zone, via CREATE state.
    """
@synceventregister
class ZoneSMDoBatchConfig(SMSyncEvent):
    """
    Initialise the zone, from batch if it is NOT DNSSEC signed.
    """
@synceventregister
class ZoneSMDoConfig(SMSyncEvent):
    """
    Initialise the zone, normally or fall back from batch.
    """
@synceventregister
class ZoneSMDoReconfig(SMSyncEvent):
    """
    Reconfigure the zone, be it DNSSEC, or related to other issues.
    """
@synceventregister
class ZoneSMDoRefresh(SMSyncEvent):
    """
    Refresh a zone, by issuing an update.
    """
@synceventregister
class ZoneSMDoSgSwap(SMSyncEvent):
    """
    Swap between sg and alt_sg a zone, transfering a zone to a new SG.
    """
@synceventregister
class ZoneSMDoSetAltSg(SMSyncEvent):
    """
    Set the alt_sg on a zone.  Doing it in SM makes it possible to do it
    live
    """
@synceventregister
class ZoneSMEnable(SMSyncEvent):
    """
    Enable zone 
    """
@synceventregister
class ZoneSMDisable(SMSyncEvent):
    """
    Disable zone 
    """
@synceventregister
class ZoneSMDelete(SMSyncEvent):
    """
    Start deleting a zone
    """
@synceventregister
class ZoneSMNukeStart(SMSyncEvent):
    """
    Start process of nuking a zone
    """
@synceventregister
class ZoneSMDoDestroy(SMSyncEvent):
    """
    Destroy a zone if its zone files are cleared
    """
@synceventregister
class ZoneSMUndelete(SMSyncEvent):
    """
    Start undeleting a zone
    """
@synceventregister
class ZoneSMEdit(SMSyncEvent):
    """
    Start a locked editing session
    """
@synceventregister
class ZoneSMEditSavedNoLock(SMSyncEvent):
    """
    Unlocked edit ZI saved, commence update
    """

@synceventregister
class ZoneSMEditExit(SMSyncEvent):
    """
    Cancel a locked editing session
    """
    def __init__(self, sm_id, edit_lock_token, *args, **kwargs):
        super().__init__(sm_id, *args, edit_lock_token=edit_lock_token,
                            **kwargs)

@synceventregister
class ZoneSMEditSaved(SMSyncEvent):
    """
    Locked Edit ZI saved, commence update
    """
    def __init__(self, sm_id, edit_lock_token, *args, **kwargs):
        super().__init__(sm_id, *args, edit_lock_token=edit_lock_token,
                            **kwargs)

@synceventregister
class ZoneSMEditLockTickle(SMSyncEvent):
    """
    Locked Edit ZI saved, commence update
    """
    def __init__(self, sm_id, edit_lock_token, *args, **kwargs):
        super().__init__(sm_id, *args, edit_lock_token=edit_lock_token,
                            **kwargs)

@eventregister
class ZoneSMEditTimeout(SMEvent):
    """
    Timeout a locked editing session
    """

@eventregister
class ZoneSMEditUpdate(SMEvent):
    """
    Update event exiting a locked edit session.
    """
    def __init__(self, sm_id, edit_lock_token, *args, **kwargs):
        super().__init__(sm_id, *args, edit_lock_token=edit_lock_token,
                            **kwargs)

def zonesmregister(class_):
    """
    Event descedant class decorator function to register class for SQL
    Alchemy mapping in init_event_class() below, called in 
    magcode.core.database.utility.setup_sqlalchemy()
    """
    sql_data['zone_sm_subclasses'].append(class_)
    return(class_)

class BaseZoneSM(StateMachine):
    """
    Base Zone State Machine.

    Parent class to contain common code between Zone SM types.
    """
    
    def __init__(self, name):
        """
        Initialise attributes etc.
        """
        # We keep domains and labels in database lowercase 
        self.name = name.lower()
        self.state = ''
        self.edit_lock = False
        self.use_apex_ns = True
        self.auto_dnssec = False
        self.nsec3 = False
        self.zone_files = False
        self.inc_updates = False
        self._reset_edit_lock
   
    def _reset_edit_lock(self):
        self.locked_by = None
        self.locked_at = None
        self.edit_lock_token = None
        self.lock_state = ZLSTATE_EDIT_UNLOCK

    def __str__(self):
        return "Zone '%s'" % self.name

    def _to_engine_timestamps(self, time_format):
        """
        Backend common function to fill out timestamps for
        to_engine methods.
        """
        if not time_format:
            locked_at = (self.locked_at.isoformat(sep=' ') 
                               if self.locked_at else None)
            deleted_start = (self.deleted_start.isoformat(sep=' ') 
                               if self.deleted_start else None)
            ctime = (self.ctime.isoformat(sep=' ') 
                               if self.ctime else None)
            mtime = (self.mtime.isoformat(sep=' ') 
                               if self.mtime else None)
        else:
            locked_at = (self.locked_at.strftime(time_format) 
                               if self.locked_at else None)
            deleted_start = (self.deleted_start.strftime(time_format) 
                               if self.deleted_start else None)
            ctime = (self.ctime.strftime(time_format) 
                               if self.ctime else None)
            mtime = (self.mtime.strftime(time_format) 
                               if self.mtime else None)
        return (locked_at, deleted_start, ctime, mtime)


    def to_engine_brief(self, time_format=None):
        """
        Brief dict of zone_sm fields for zone engine
        """
        locked_at, deleted_start, ctime, mtime \
                = self._to_engine_timestamps(time_format)
        result =  {'zone_id': self.id_, 'zi_id': self.zi_id,
                'name': self.name, 'state': self.state, 
                'soa_serial': self.soa_serial, 
                'ctime': ctime, 'mtime': mtime,
                'deleted_start': deleted_start}
        result['reference'] = (self.reference.reference if self.reference
                                    else None)
        return result

    def to_engine(self, time_format=None):
        """
        Return dict of zone_sm fields for zone engine
        """
        locked_at, deleted_start, ctime, mtime \
                = self._to_engine_timestamps(time_format)
        result =  {'zone_id': self.id_, 'zi_id': self.zi_id, 
                'zi_candidate_id': self.zi_candidate_id,
                'state': self.state, 'soa_serial': self.soa_serial,
                'zone_type': self.zone_type, 'name': self.name, 
                'lock_state': self.lock_state,
                'locked_by': self.locked_by,
                'locked_at': locked_at,
                'use_apex_ns': self.use_apex_ns, 'edit_lock': self.edit_lock,
                'auto_dnssec': self.auto_dnssec, 'nsec3': self.nsec3,
                'edit_lock_token': self.edit_lock_token,
                'inc_updates': self.inc_updates,
                'sg_name': self.sg.name,
                'deleted_start': deleted_start,
                'ctime': ctime, 'mtime': mtime, }
        result['reference'] = (self.reference.reference if self.reference
                                    else None)
        result['alt_sg_name'] = (self.alt_sg.name if self.alt_sg 
                                    else None)
        return result

@smregister
@typeregister
class ZoneSM(BaseZoneSM):
    """
    IntermediateZone State Machine class which defines SQL Alchemy
    Accessors.
    """
    _sm_events = ()
    # self._tmplate_names indirection used so that config file settings
    # will take effect.
    _template_names = ()

    @classmethod
    def _mapper_properties(class_):
        zi_type = sql_types['ZoneInstance']
        zone_sectag_type = sql_types['ZoneSecTag']
        sg_type = sql_types['ServerGroup']
        reference_type = sql_types['Reference']
        zi_table = sql_data['tables'][zi_type]
        zone_sm_table = sql_data['tables'][ZoneSM]
        sg_table = sql_data['tables'][sg_type]
        ug_type = sql_types['UpdateGroup']
        ug_table = sql_data['tables'][ug_type]
        rn_type = sql_types['ReverseNetwork']
        return {'all_zis': relationship(zi_type,
                            primaryjoin=(zi_table.c.zone_id
                                == zone_sm_table.c.get('id')),
                            lazy='dynamic', passive_deletes=True),
                'zi':   relationship(zi_type,
                            primaryjoin=(zi_table.c.get('id')
                                == zone_sm_table.c.zi_id),
                            viewonly=True),
                'zi_candidate':   relationship(zi_type,
                            primaryjoin=(zi_table.c.get('id')
                                == zone_sm_table.c.zi_candidate_id),
                            foreign_keys=[zone_sm_table.c.zi_candidate_id],
                            viewonly=True),
                'sg':  relationship(sg_type, primaryjoin=(
                            sg_table.c.get('id') == zone_sm_table.c.sg_id),
                            viewonly=True),
                'alt_sg':  relationship(sg_type, primaryjoin=(
                        sg_table.c.get('id') == zone_sm_table.c.alt_sg_id),
                            viewonly=True),
                'reference': relationship(reference_type, viewonly=True),
                'sectags': relationship(zone_sectag_type, passive_deletes=True),
                'update_groups': relationship(ug_type, passive_deletes=True,
                                        order_by=ug_table.c.get('id'), 
                                        backref='zone'),
                'reverse_network': relationship(rn_type, passive_deletes=True,
                                uselist=False, backref='zone'),
                            }


    @classmethod
    def sa_map_subclass(class_):
        sql_data['mappers'][class_] = mapper(class_,
                    inherits=sql_data['mappers'][ZoneSM], 
                    polymorphic_identity=class_.__name__)
        sql_data['zone_sm_type_list'].append(class_.__name__)

    def add_sectag(self, db_session, zone_sectag):
        """
        Add a security tag to the Zone
        """
        ZoneSecTag = sql_types['ZoneSecTag']
        # Skip admin sectag
        if zone_sectag == ZoneSecTag(settings['admin_sectag']):
            return
        # Make sure sec tag does not already exist
        if zone_sectag in self.sectags:
            return
        db_session.add(zone_sectag)
        self.sectags.append(zone_sectag)

    def remove_sectag(self, db_session, zone_sectag):
        """
        Remove a security tag from the zone
        """
        ZoneSecTag = sql_types['ZoneSecTag']
        # Skip admin sectag
        if zone_sectag == ZoneSecTag(settings['admin_sectag']):
            return
        # form list of objects to delete
        del_list = [x for x in self.sectags if x == zone_sectag]
        for x in del_list:
            self.sectags.remove(x)
        while del_list:
            db_session.delete(del_list[-1])
            x = del_list.pop()
            if not x:
                continue
            del x

    def remove_all_sectags(self, db_session):
        """
        Remove all security tags from the zone.
        """
        while self.sectags:
            sectag = self.sectags.pop()
            if not sectag:
                continue
            db_session.delete(sectag)
            del sectag

    def copy_zone_sectags(self, db_session, src_zone_sm):
        """
        add sectag list to a zone
        """
        ZoneSecTag = sql_types['ZoneSecTag']
        admin_sectag = ZoneSecTag(settings['admin_sectag'])
        for src_sectag in src_zone_sm.sectags:
        # Skip admin sectag
            if src_sectag == admin_sectag:
                continue
            zone_sectag = ZoneSecTag(src_sectag.sectag)
            self.add_sectag(db_session, zone_sectag)

    def replace_all_sectags(self, db_session, *zone_sectags):
        """
        Replace all sectags for a zone
        """
        self.remove_all_sectags(db_session)
        for zone_sectag in zone_sectags:
            self.add_sectag(db_session, zone_sectag)

    def list_sectags(self, db_session):
        """
        List all security tags for this zone as JSON
        """
        result = [sql_types['ZoneSecTag'](settings['admin_sectag'])\
                        .to_engine_brief()]
        result.extend([t.to_engine_brief() for t in self.sectags])
        return result

    def set_sg(self, sg):
        """
        Set the server group this zone is served from.
        """
        if hasattr(self, 'sg') and self.sg:
            old_sg = self.sg
            old_sg.zones.remove(self)
            self.sg = None
        sg.zones.append(self)
        self.sg = sg

    def _set_alt_sg(self, sg):
        """
        Set the alternate server group this zone is served from.
        """
        if hasattr(self, 'alt_sg') and self.alt_sg:
            old_sg = self.alt_sg
            old_sg.alt_zones.remove(self)
            self.alt_sg = None
        if sg:
            sg.alt_zones.append(self)
        self.alt_sg = sg

    def write_config(self, include_file, db_session, server_acls, 
            replica_sg=None):
        """
        Fill out master server configuration template

        Stub function that needs overriding in descendant class
        """
        raise IOError(errno.EINVAL, os.strerror(errno.EINVAL), include_name)
        # Exceptions for this caught by an outer try: in the 
        # self._tmplate_names indirection used so that config file settings
        # will take effect.
        #template_name = (settings['master_template_dir'] + '/' 
        #                        + settings[self._template_names[0]])
        #template = read_template(template_name)
        #filler = { 'name': self.name, }
        #section = template % filler
        # include_file.write(section)
    
    def is_disabled_or_deleted(self):
        """
        Test to see if a zone is disabled or deleted.

        Saves having to import ZSTATE_DISABLED and thus import nesting...
        """
        return self.state in (ZSTATE_DISABLED, ZSTATE_DELETED,)

    @classmethod
    def query_is_not_disabled_deleted(self, query):
        """
        Test to see if a zone is not disabled or deleted.

        Saves having to import ZSTATE_DISABLED and thus import nesting...
        """
        return query.filter(and_(self.state != ZSTATE_DISABLED,
                    self.state != ZSTATE_DELETED))

    def is_not_configured(self):
        """
        Test to see if a zone is not configured.

        Saves having to import ZSTATE_DISABLED and thus import nesting...
        """
        return self.state in (ZSTATE_DISABLED, ZSTATE_DELETED, ZSTATE_CREATE,
                            ZSTATE_BATCH_CREATE_1)
    
    @classmethod
    def query_sg_is_configured(self, query):
        """
        Test to see if a zone is configured for use in server config

        Considers ZSTATE_RESET to be a configured state, so that zones are
        not removed from servers during a ZonsSM reset
        Saves having to import ZSTATE_DISABLED and thus import nesting...
        """
        return query.filter(~ self.state.in_((ZSTATE_DISABLED, ZSTATE_DELETED,
            ZSTATE_CREATE, ZSTATE_BATCH_CREATE_1)) )

    @classmethod
    def query_is_configured(self, query):
        """
        Test to see if a zone is configured.

        Saves having to import ZSTATE_DISABLED and thus import nesting...
        """
        return query.filter(~ self.state.in_((ZSTATE_DISABLED, ZSTATE_DELETED,
            ZSTATE_CREATE, ZSTATE_RESET, ZSTATE_BATCH_CREATE_1)) )

    def is_deleted(self):
        """
        Test to see if a zone is deleted.

        Saves having to import ZSTATE_DELETED and thus import nesting...
        """
        return self.state in (ZSTATE_DELETED,)
    
    @classmethod
    def query_is_not_deleted(self, query):
        """
        Add a test to a query to see if a zone is deleted or not
        """
        return query.filter(self.state != ZSTATE_DELETED)

    def is_disabled(self):
        """
        Test to see if a zone is disabled.

        Saves having to import ZSTATE_DISABLED and thus import nesting...
        """
        return self.state == ZSTATE_DISABLED

    @classmethod
    def query_inc_updates(self, query):
        """
        Add a test to a query to see if a zone has inc_updates enabled.
        """
        return query.filter(self.inc_updates == True)

    def _do_incremental_updates(self, db_session, zi,
                        process_inc_updates=False):
        """
        Process incremental updates for a zone
        """
        # Check if zone edit locked, if so, defer
        # This to prevent updates being 'lost'
        if (not process_inc_updates and self.edit_lock == ZLSTATE_EDIT_LOCK):
            return zi
        # Pull in all update groups for zone
        UpdateGroup = sql_types['UpdateGroup']
        query = db_session.query(UpdateGroup)\
                .filter(UpdateGroup.zone_id == self.id_)
        updates = query.all()
        # Check to see if there are updates
        if not len(updates):
            return zi
        # Check and see if updates are all PTR related, if so, follow different
        # copy algorithm
        normal_updates = [ug for ug in updates if not ug.ptr_only]
        if not len(normal_updates):
            # only create new candidate ZI only if config_hold_time has passed
            # This means published reverse ZI gets morphed, only time a
            # ZI contents get changed other than Apex records and Zone 
            # TTL updates
            time = db_time(db_session)
            freeze_time = timedelta(
                            minutes=float(settings['master_hold_timeout']))
            if (self.zi and time > (self.zi.ctime + freeze_time) 
                and self.zi_candidate_id == self.zi_id):
                zi = zi.copy(db_session)
            elif self.zi_candidate_id == self.zi_id:
                log_debug("Zone '%s' - republishing old reverse ZI %s."
                            % (self.name, zi.id_))
       
        # Normal copy algorithm
        elif (self.zi_candidate_id == self.zi_id):
            # create new candidate ZI
            zi = zi.copy(db_session)

        # Apply each update group to candidate ZI
        for ug in updates:
            zi.exec_update_group(db_session, ug)

        # return updated zi
        return zi
        

@smregister
@zonesmregister
class StaticZoneSM(ZoneSM):
    """
    Static Zone File State Machine

    Implements the traditional static zone file
    Currently just a place holder
    """
    pass
    _template_names = (MASTER_STATIC_TEMPLATE,)


@smregister
@zonesmregister
class SlaveZoneSM(ZoneSM):
    """
    Slave Zone File State Machine

    Implements a slaved master
    Currently just a place holder
    """
    _template_names = (MASTER_SLAVE_TEMPLATE,)
    pass


@smregister
@zonesmregister
class DynDNSZoneSM(ZoneSM):
    """
    Dynamic DNS Zone State Machine

    Implements Zone State machine that uses Dynamic DNS for Updates
    """
    # self._tmplate_names indirection used so that config file settings
    # will take effect.
    _template_names = (MASTER_DYNDNS_TEMPLATE, MASTER_AUTO_DNSSEC_TEMPLATE,)
    _sm_events= (ZoneSMDoReset, ZoneSMDoSgSwap, ZoneSMDoSetAltSg,
            ZoneSMDoBatchConfig, ZoneSMBatchConfig, ZoneSMReconfigCheck,
            ZoneSMDoReconfig, ZoneSMReconfigUpdate, ZoneSMDoConfig, 
            ZoneSMConfig, ZoneSMRemoveZoneFiles,
            ZoneSMEnable, ZoneSMDisable, ZoneSMDoRefresh, ZoneSMEditExit,
            ZoneSMEdit, ZoneSMUpdate, ZoneSMEditUpdate, ZoneSMEditSaved,
            ZoneSMEditSavedNoLock,
            ZoneSMEditLockTickle, ZoneSMEditTimeout, ZoneSMDelete,
            ZoneSMUndelete, ZoneSMNukeStart, ZoneSMDoDestroy)

    def __init__(self, name, edit_lock=False, use_apex_ns=True,
            auto_dnssec=False, nsec3=False, inc_updates=False):
        """
        Initialise Zone SM
        """
        super().__init__(name)
        self.edit_lock = edit_lock
        self.use_apex_ns = use_apex_ns
        self.auto_dnssec = auto_dnssec
        self.nsec3 = nsec3
        self.inc_updates = inc_updates
        self.state = ZSTATE_INIT
    
    def _process_edit_lock_token_mismatch(self):
        """
        Handle a lock token mismatch.  Needed to handle lock timeout as
        well as actually locked zone.
        """
        if self.lock_state == ZLSTATE_EDIT_UNLOCK:
            raise ZoneEditLockTimedOut(self.name)
        else:
            raise ZoneEditLocked(self.name, self.edit_lock_token,
                    self.locked_by, self.locked_at)

    def write_zone_file(self, db_session, op_exc):
        """
        Write out zone file.

        Usable from outside as root for recovery purposes
        """
        # Get zi, if not found, fail gracefully
        zi_type = sql_types['ZoneInstance']
        err_string = ''
        try:
            zi = db_session.query(zi_type)\
                    .filter(zi_type.id_ == self.zi_candidate_id).one()
        except NoResultFound as exc:
            err_string = str(exc)
        if err_string:
            raise op_exc("Zone '%s' - %s" % (self.name, err_string))
        # Write/overwrite current zi to NS dynamic dir
        err_string = ''
        err_filename = ''
        try:
            dynamic_zone_dir = settings['master_dyndns_dir']
            # Remove dot at end of zone name as this gives more
            # human literate filenames
            human_name = self.name[:-1] if self.name.endswith('.') \
                                    else self.name
            zone_file = settings['master_dyndns_dir'] + '/' + human_name
            zone_file_jnl = zone_file + '.jnl'
            zi_data = zi.to_data(all_rrs=True)
            prefix = '.' + os.path.basename(human_name) + '-'
            (fd, tmp_filename) = mkstemp(
                        dir=dynamic_zone_dir,
                        prefix=prefix)
            tmp_file = io.open(fd, mode='wt')
            reference = self.reference.reference if self.reference else None
            data_to_bind(zi_data, self.name, tmp_file, reference=reference,
                        for_bind=True)
            tmp_file.close()
            # Rename tmp file into place so that replacement is atomic
            try:
                uid = pwd.getpwnam(settings['run_as_user']).pw_uid
                bind_gid = grp.getgrnam(settings['zone_file_group']).gr_gid
            except KeyError as exc:
                msg = ("Could not look up group '%s' or user '%s'"
                        " for zone file for %s" 
                        % (settings['zone_file_group'],
                            settings['run_as_user'],
                            self.name))
                raise op_exc(msg)
            os.chown(tmp_filename, uid, bind_gid)
            os.chmod(tmp_filename, int(settings['zone_file_mode'],8))
            # Update zone_files flag
            self.zone_files = True
            try:
                # Remove journal file - can cause trouble with bind
                os.unlink(zone_file_jnl)
            except:
                pass
            os.rename(tmp_filename, zone_file)
        except (IOError, OSError) as exc:
            err_string = exc.strerror
            err_filename = exc.filename
        finally:
            # clean up if possible
            try:
                os.unlink(tmp_filename)
            except:
                pass
        if err_string:
            msg = ( "Could not write file '%s' - %s." 
                            % (err_filename, err_string))
            raise op_exc(msg)

    def _remove_zone_files(self, event):
        """
        Tidy up routine to remove zone files.  Does not matter it if errors
        """
        db_session = event.db_session
        # Set self.zone_files false as this janitor event for zone is being
        # executed. If another zone active, when it is deleted or disabled it
        # will remove the files if another zone instance is not active then...
        try:
            query = db_session.query(ZoneSM)\
                    .filter(ZoneSM.name == self.name)\
                    .filter(and_(ZoneSM.state != ZSTATE_DELETED, 
                                ZoneSM.state != ZSTATE_DISABLED))
            result = query.all()
            if result:
                self.zone_files = False
                return (RCODE_OK, "Zone '%s' - not removing files "
                        "as another zone instance active" 
                        % self.name)
        except NoResultFound:
            pass
        finally:
            del result
        human_name = self.name[:-1] if self.name.endswith('.') \
                                    else self.name
        zone_file = settings['master_dyndns_dir'] + '/' + human_name
        zone_file_jnl = zone_file + '.jnl'
        try:
            os.unlink(zone_file)
            os.unlink(zone_file_jnl)
        except:
            pass
        self.zone_files = False
        return (RCODE_OK, "Zone '%s' - tidy up - zone files probably deleted" 
                        % self.name)

    def _do_destroy(self, event):
        """
        ZoneSM routine to remove zone, will only succeed if zone files have 
        already been removed, otherwise will queue a ZoneSMRemoveZoneFiles
        """
        if not self.zone_files:
            event.db_session.delete(self)
            return (RCODE_OK, 
                    "Zone '%s' - destroying, zone files deleted"
                    % self.name)
        # Can't do it now, but need to seed our own future destruction
        # Use coalesce_period to avoid double events if possible
        buffer_period = timedelta(seconds=3*float(settings['sleep_time']))
        coalesce_period = timedelta(
                            minutes=2*float(settings['master_hold_timeout']))
        create_event(ZoneSMRemoveZoneFiles, db_session=event.db_session, 
                        sm_id=self.id_, zone_id=self.id_, name=self.name,
                        coalesce_period=coalesce_period,
                        delay=coalesce_period+buffer_period)
        raise ZoneNotDestroyedFilesExist(
                "Zone '%s' - not destroyed zone files still exist" 
                                            % self.name)

    def _do_batch_config(self, event):
        """
        Batch configure zone
        """
        # Add zi from event if zone is being created
        self.zi_candidate_id = event.py_parameters['zi_id']
        # Initialise self.zi_id so that show_zone works on zone creation 
        if not self.zi_id:
            self.zi_id = event.py_parameters['zi_id']
        self.state = ZSTATE_BATCH_CREATE_1
        create_event(ZoneSMBatchConfig, db_session=event.db_session,
                        sm_id=self.id_, zone_id=self.id_, name=self.name)
        return (RCODE_OK, "Zone '%s' - initialising" % self.name)

    def _batch_config(self, event):
        """
        Initialise zone configuration to named.conf on master and servers
        """
        self.write_zone_file(event.db_session, BatchCreateFailed)
        self.state = ZSTATE_BATCH_CREATE_2
        # Queue ZoneSMReconfigCheck
        zone_sm_reconfig_schedule(event.db_session, self, ZoneSMReconfigCheck)
        return (RCODE_OK, "Zone '%s' -  Wrote seed zone file, queuing reconfig" 
                        % self.name)

    def _reconfig_check(self, event):
        """
        Check that zone is loaded on zone creation
        """
        db_session = event.db_session
        # Run update engine
        (rcode, msg, soa_serial) = update_engine['dyndns']\
                                    .read_soa(self.name)

        # Handle auto reset of Zone SM if DNS server is not configured
        if rcode == RCODE_RESET:
            log_info(msg)
            msg = "reseting ZoneSM as server not configured" 
            return self._do_reset(event, msg, randomize=True, via_create=True)

        if rcode == RCODE_ERROR:
            raise ReadSoaRetry(msg)
        elif rcode == RCODE_FATAL:
            raise ReadSoaFailed(msg)
        if self.zi_id != self.zi_candidate_id:
            self.zi_id = self.zi_candidate_id
        self.state = ZSTATE_PUBLISHED
        return (RCODE_OK, "Zone '%s' - Master NS loaded successfully"
                            % self.name)

    def _retry_reconfig(self, event, randomize=False):
        """
        Retry ZoneSMReconfigUpdate
        """
        self.state = ZSTATE_UNCONFIG
        # Queue ZoneSMReconfigUpdate
        zone_sm_reconfig_schedule(event.db_session, self, ZoneSMReconfigUpdate, 
                            randomize=randomize)
        msg = "Retrying reconfig of ZoneSM as server not configured" 
        return (RCODE_OK, msg)

    def _do_reset(self, event, msg=None, randomize=False, via_create=False):
        """
        Reset the Zone SM, going via CREATE state as per normal zone creation 
        """
        db_session = event.db_session
        # Specifically fetch parameters
        zi_id = event.py_parameters.get('zi_id')
        if zi_id:
            self.zi_candidate_id = event.py_parameters['zi_id']
            # Initialise self.zi_id so that show_zone works on zone creation 
            if not self.zi_id:
                self.zi_id = event.py_parameters['zi_id']
        self.state = ZSTATE_CREATE if via_create else ZSTATE_RESET
        # Queue ZoneSMConfig, only master bind reconfiguration,
        # to preserve anything being currently served on servers
        zone_sm_reconfig_schedule(db_session, self, ZoneSMConfig, 
                                    master_reconfig=True, randomize=randomize)
        msg = "Zone '%s' - reseting SM" % self.name
        return (RCODE_OK, msg)

    def _do_config(self, event):
        """
        Initialise zone creation normally 
        """
        db_session = event.db_session
        # Add zi from event if zone is being created
        self.zi_candidate_id = event.py_parameters['zi_id']
        # Initialise self.zi_id so that show_zone works on zone creation 
        if not self.zi_id:
            self.zi_id = event.py_parameters['zi_id']
        self.state = ZSTATE_CREATE
        create_event(ZoneSMConfig, db_session=db_session,
                        sm_id=self.id_, zone_id=self.id_, name=self.name)
        return (RCODE_OK, "Zone '%s' - initialising" % self.name)

    def _config(self, event, write_file_exc=CreateFailed):
        """
        Add configuration to named.conf on master and servers
        """
        db_session = event.db_session

        # Run update engine - we are checking to see if zone is in bind
        (rcode, msg, soa_serial) = update_engine['dyndns']\
                                    .read_soa(self.name)
        if rcode == RCODE_ERROR:
            raise ReadSoaRetry(msg)
        if rcode == RCODE_OK:
            msg = ("Zone '%s' - reconfiguring ZoneSM as server not configured"
                    % self.name)
            log_info(msg)
            return self._do_reset(event, via_create=False)

        self.write_zone_file(db_session, write_file_exc)
        self.state = ZSTATE_UNCONFIG
        # Queue ZoneSMReconfigUpdate
        zone_sm_reconfig_schedule(db_session, self, ZoneSMReconfigUpdate)
        return (RCODE_OK, "Zone '%s' -  Wrote seed zone file, queuing reconfig" 
                        % self.name)
    
    def _reconfig_update(self,event):
        """
        Update upon rndc reconfig
        """
        return self._update(event)

    def _nuke_start(self, event):
        """
        Prepare a zone to be nuked by setting it to deleted state.
        """
        return self._delete(event, nuke_start=True)

    def _delete(self, event, nuke_start=False):
        """
        Delete processing for zone
        """
        # check that zone is not EDIT_LOCKED
        if  (self.lock_state == ZLSTATE_EDIT_LOCK):
            raise ZoneEditLocked(self.name, self.edit_lock_token,
                    self.locked_by, self.locked_at)
        # set state to DELETED
        if self.state in (ZSTATE_PUBLISHED, ZSTATE_UPDATE):
            if self.zi_id != self.zi_candidate_id:
                self.zi_id = self.zi_candidate_id
        self.state = ZSTATE_DELETED
        if nuke_start:
             self.deleted_start = None
        else:
            self.deleted_start = db_time(event.db_session)
        # Queue MasterSMPartialReconfig
        zone_sm_reconfig_schedule(event.db_session, self, ZoneSMRemoveZoneFiles)
        if nuke_start:
            msg = ("Zone '%s' - DELETED state preparing to nuke"
                            % self.name)
        else:
            msg = ("Zone '%s' - going into DELETED state"
                            % self.name)
        return (RCODE_OK, msg)

    def _event_failure(self, event):
        """
        Fail an event when zone is in INIT transient state.
        """
        raise ZoneSMEventFailure(
                "Zone '%s' - event failed as zone in transient INIT state"
                % self.name)

    def _update_fail(self, event):
        """
        Fail an when zone is in UPDATE state with edit_locked changes being
        saved.
        """
        raise ZoneSMUpdateFailure(
                "Zone '%s' - Failure due to UPDATE state changes being saved"
                % self.name)

    def _undelete(self, event):
        """
        Undelete a zone from DELETED to UNCONFIG, then proceed to enable it.
        """
        # Get zi, if not found, fail gracefully
        zi_type = sql_types['ZoneInstance']
        err_string = ''
        try:
            zi = event.db_session.query(zi_type)\
                    .filter(zi_type.id_ == self.zi_candidate_id).one()
        except NoResultFound as exc:
            err_string = str(exc)
        if err_string:
            raise ZoneSMUndeleteFailure("Zone '%s' - %s" % (self.name, msg))
        # Check that we are only zone instance
        db_session = event.db_session
        try:
            query = db_session.query(ZoneSM)\
                    .filter(ZoneSM.name == self.name)\
                    .filter(ZoneSM.state != ZSTATE_DELETED)
            result = query.all()
            if result:
                raise ZoneSMUndeleteFailure(
                    "Zone '%s' - Failure as other instances of zone are active"
                        % self.name)
        except NoResultFound:
            pass
        # Go to CREATE
        self.deleted_start = None
        self.zi_candidate_id = self.zi_id
        self.state = ZSTATE_CREATE
        create_event(ZoneSMConfig, db_session=event.db_session,
                        sm_id=self.id_, zone_id=self.id_, name=self.name)
        return (RCODE_OK, "Zone '%s' - undeleted and initialising" % self.name)

    def _disable(self, event):
        """
        Remove configuration to named.conf on master and servers
        """
        if self.state in (ZSTATE_PUBLISHED, ZSTATE_UPDATE):
            if self.zi_id != self.zi_candidate_id:
                self.zi_id = self.zi_candidate_id
        self.state = ZSTATE_DISABLED
        # Execute Configuration SM here
        # Do a MasterSMPartialReconfig here
        zone_sm_reconfig_schedule(event.db_session, self, ZoneSMRemoveZoneFiles)
        return (RCODE_OK, "Zone '%s' - disabling" % self.name)

    def _already_disabled(self, event):
        """
        Zone already disabled
        """
        raise ZoneAlreadyDisabled("Zone '%s' - already disabled" % self.name)

    def _enable(self, event):
        """
        Proceed to enable zone
        """
        self.state = ZSTATE_CREATE
        create_event(ZoneSMConfig, db_session=event.db_session,
                        sm_id=self.id_, zone_id=self.id_, name=self.name)
        return (RCODE_OK, "Zone '%s' - enabled and initialising" % self.name)

    def _already_enabled(self, event):
        """
        Zone already enabled
        """
        raise ZoneAlreadyEnabled("Zone '%s' - already enabled" % self.name)

    def _reset_disabled(self, event):
        """
        Zone disabled - can't reset
        """
        raise ZoneResetDisabled("Zone '%s' - disabled, can't reset" % self.name)

    def _do_reconfig(self, event):
        """
        Reconfigure zone
        """
        if self.state in (ZSTATE_PUBLISHED, ZSTATE_UPDATE):
            if self.zi_id != self.zi_candidate_id:
                self.zi_id = self.zi_candidate_id
        self.state = ZSTATE_UNCONFIG
        zone_sm_reconfig_schedule(event.db_session, self, ZoneSMReconfigUpdate)
        return (RCODE_OK, "Zone '%s' - reconfiguring" % self.name)

    def _do_refresh(self, event):
        """
        Refresh a zone by queuing an update event.
        """
        if self.state in (ZSTATE_PUBLISHED,):
            zi_candidate_id = event.py_parameters.get('zi_id')
            wrap_soa_serial = event.py_parameters.get('wrap_soa_serial')
            candidate_soa_serial = event.py_parameters.get(
                                                    'candidate_soa_serial')
            force_soa_serial_update = event.py_parameters.get(
                                                    'force_soa_serial_update')
            if (wrap_soa_serial or candidate_soa_serial 
                                            or force_soa_serial_update):
                # Can only do these operations if zone is not locked.
                if (self.lock_state == ZLSTATE_EDIT_LOCK):
                    raise ZoneEditLocked(self.name, self.edit_lock_token,
                            self.locked_by, self.locked_at)
            if zi_candidate_id:
                if (self.lock_state == ZLSTATE_EDIT_LOCK):
                    raise ZoneEditLocked(self.name, self.edit_lock_token,
                            self.locked_by, self.locked_at)
                self.zi_candidate_id = zi_candidate_id
            elif self.zi_id != self.zi_candidate_id:
                self.zi_id = self.zi_candidate_id
            create_event(ZoneSMUpdate, db_session=event.db_session,
                     sm_id=self.id_, zone_id=self.id_, name=self.name,
                     publish_zi_id=self.zi_candidate_id,
                     wrap_soa_serial=wrap_soa_serial,
                     candidate_soa_serial=candidate_soa_serial,
                     force_soa_serial_update=force_soa_serial_update)
        return (RCODE_OK, "Zone '%s' - refreshing" % self.name)

    def _do_sg_swap(self, event):
        """
        Swap SGs, and then reconfig refresh a zone by queuing events.
        """
        db_session = event.db_session
        if not self.alt_sg:
            raise ZoneNoAltSg("Zone '%s' - no alt_sg, swap failed." 
                                                        % self.name)
        new_sg = self.alt_sg
        new_alt_sg = self.sg
        self.set_sg(new_sg)
        self._set_alt_sg(new_alt_sg)
        zone_sm_reconfig_schedule(db_session, self, ZoneSMUpdate, 
                randomize=True, publish_zi_id=self.zi_candidate_id)
        return (RCODE_OK, "Zone '%s' - SG reconfig and then refresh" 
                        % self.name)

    def _do_set_alt_sg(self, event):
        """
        Set alt SG.  Needed to lock Zone to update zone_sm row in DB.
        """
        db_session = event.db_session
        alt_sg_name = event.py_parameters['alt_sg_name']
        if alt_sg_name:
            alt_sg = find_sg_byname(db_session, alt_sg_name, raise_exc=False)
            if not alt_sg:
                raise ZoneNoAltSg("Zone '%s' - no alt_sg %s, set failed." 
                                                        % (self.name, sg_name))
            reconf_sg = alt_sg
        else:
            alt_sg = None
            reconf_sg = self.alt_sg
        self._set_alt_sg(alt_sg)
        reconfig_sg(db_session, reconf_sg.id_, reconf_sg.name)
        return (RCODE_OK, "Zone '%s' - alt SG set and SG reconfig queued" 
                        % self.name)

    def _edit(self, event):
        """
        Edit zone configuration.
        """
        db_session = event.db_session
        if (not self.edit_lock):
            return (RCODE_NOCHANGE, "Edit locking turned off for zone '%s'" 
                        % self.name)
        if  (self.lock_state == ZLSTATE_EDIT_LOCK):
            raise ZoneEditLocked(self.name, self.edit_lock_token,
                    self.locked_by, self.locked_at)

        timeout = timedelta(
                        minutes=float(settings['edit_lock_timeout']))
        if timeout: 
            timeout_event = create_event(ZoneSMEditTimeout,
                                db_session=db_session,
                                delay=timeout, sm_id=self.id_, 
                                zone_id=self.id_, name=self.name)
            lock_id = timeout_event.id_
        else:
            lock_id = event.id_
        self.edit_lock_token = lock_id
        self.lock_state = ZLSTATE_EDIT_LOCK
        self.locked_by = event.py_parameters.get('locked_by') 
        self.locked_at = db_time(db_session)
        event.py_results['edit_lock_token'] =  self.edit_lock_token
        return (RCODE_OK, "Locked edit zone '%s' entered - token '%s'" 
                % (self.name, self.edit_lock_token))

    def _edit_locked(self, event):
        """
        Attempt to edit  a zone while edit locked!!

        raise and bomb...
        """
        raise ZoneEditLocked(self.name, self.edit_lock_token, self.locked_by, 
                            self.locked_at)

    def _edit_not_locked(self, event):
        """
        Attempt to update a zone while not edit locked!!

        raise and bomb...
        """
        raise ZoneNotEditLocked(self.name)

    def _edit_exit(self, event):
        """
        Exit edit lock state.  Need value of lock token to exit.
        """
        if not self.edit_lock_token:
            if self.lock_state != ZLSTATE_EDIT_UNLOCK:
                raise ZoneSMLockFailure(
                        "Zone '%s' - lock mechanism in bad state"
                        % self.name)
            return (RCODE_OK, "Zone '%s' - exit edit, zone not locked"
                    % self.name) 
        if  (event.py_parameters['edit_lock_token'] != self.edit_lock_token):
            raise ZoneEditLocked(self.name, self.edit_lock_token,
                    self.locked_by, self.locked_at)
        # This will cancel any existing timeout events with this
        # edit_lock_token.  If it is a ZoneSMEditEvent, it is the one
        # processed to begin this session.
        cancel_event(self.edit_lock_token, db_session=event.db_session)
        old_lock_state = self.lock_state
        self._reset_edit_lock()
        return (RCODE_OK, "Exiting %s for zone '%s'" % (old_lock_state,
                                                            self.name))
    
    def _update_edit_exit(self, event):
        """
        Clear edit lock in UPDATE state by queuing ZoneSMEditUpdate event
        """
        if  (event.py_parameters['edit_lock_token'] != self.edit_lock_token):
            raise ZoneEditLocked(self.name, self.edit_lock_token, 
                    self.locked_by, self.locked_at)
        create_event(ZoneSMEditUpdate, db_session=event.db_session,
                     sm_id=self.id_, zone_id=self.id_, name=self.name,
                     publish_zi_id=self.zi_candidate_id,
                     edit_lock_token=event.py_parameters['edit_lock_token'])
        return (RCODE_OK, "Zone '%s' - exiting lock by queuing"
                " ZoneSMEditUpdate" % self.name)

    def _edit_lock_tickle(self, event):
        """
        Tickle edit lock timeout  Need correct value of lock token to execute.
        """
        if  (event.py_parameters['edit_lock_token'] != self.edit_lock_token):
            self._process_edit_lock_token_mismatch()
        timeout = timedelta(
                        minutes=float(settings['edit_lock_timeout']))
        if not timeout:
            return (RCODE_NOEFFECT, "Edit Lock Timeout disabled")
        reschedule_event(self.edit_lock_token, db_session=event.db_session,
                delay=timeout)
        return (RCODE_OK, "Tickled timeout for zone '%s'" % self.name)

    def _edit_timeout(self, event):
        """
        Timeout edit lock state.  Need value of lock token to do timeout.
        """
        old_lock_state = self.lock_state
        self._reset_edit_lock()
        return (RCODE_OK, "Exiting %s for zone '%s'" % (old_lock_state,
                                            self.name))

    def _edit_saved_no_lock(self, event):
        """
        Unlocked edit saved, puts zi in place, issues a publish event
        """
        db_session = event.db_session
        zi_candidate_id = event.py_parameters['zi_id']
        self.zi_candidate_id = zi_candidate_id
        event = ZoneSMUpdate(sm_id=self.id_, zone_id=self.id_, name=self.name,
                                publish_zi_id=zi_candidate_id)
        # This call does a db_session.commit()
        queue_event(event, db_session=db_session, commit=True,
                    signal_queue_daemon=True)
        return (RCODE_OK, "Zone '%s' - unlocked edit saved,"
                        " ZoneSMUpdate queued" % self.name)

    def _other_edit_saved_no_lock(self, event):
        """
        Unlocked edit saved, puts zi in place, in non-PUBLISHED state
        """
        db_session = event.db_session
        zi_id = event.py_parameters['zi_id']
        self.zi_candidate_id = zi_id
        self.zi_id = zi_id
        return (RCODE_OK, "Zone '%s' - unlocked edit saved" % self.name)


    def _edit_saved_to_update(self, event):
        """
        Lock edit saved, commence update state
        """
        if  (event.py_parameters['edit_lock_token'] != self.edit_lock_token):
            self._process_edit_lock_token_mismatch()
        # This will cancel any existing timeout events with this
        # edit_lock_token.  If it is a ZoneSMEditEvent, it is the one
        # processed to begin this session.
        cancel_event(self.edit_lock_token, db_session=event.db_session)
        old_state = self.state
        self.state = ZSTATE_UPDATE
        self.zi_candidate_id = event.py_parameters['zi_id']
        create_event(ZoneSMEditUpdate, db_session=event.db_session,
                     sm_id=self.id_, zone_id=self.id_, name=self.name,
                     publish_zi_id=self.zi_candidate_id,
                     edit_lock_token=event.py_parameters['edit_lock_token'])
        return (RCODE_OK, "Exiting  %s for zone '%s' - edit saved, updating" 
                    % (old_state, self.name))

    def _other_edit_saved(self, event):
        """
        Do edit saved for states other than PUBLISHED
        """
        if  (event.py_parameters['edit_lock_token'] != self.edit_lock_token):
            self._process_edit_lock_token_mismatch()
        # This will cancel any existing timeout events with this
        # edit_lock_token.  If it is a ZoneSMEditEvent, it is the one
        # processed to begin this session.
        cancel_event(self.edit_lock_token, db_session=event.db_session)
        self._reset_edit_lock()
        zi_id = event.py_parameters['zi_id']
        self.zi_candidate_id = zi_id
        self.zi_id = zi_id
        return (RCODE_OK, "Zone '%s' - edit saved, updated" 
                    % self.name)

    def _update_dnssec_preprocess(self):
        """
        Pre update processing for DNSSEC
        """
        dnssec_args = {}
        if self.auto_dnssec and not self._check_dnssec_keys():
            log_error ("Zone '%s' - DNSSEC keys are not present." % self.name)
            return {}
        if not self.auto_dnssec:
            dnssec_args['clear_dnskey'] = True
        elif self.auto_dnssec:
            if self.nsec3:
                dnssec_args['nsec3_seed'] = True
            elif not self.nsec3:
                dnssec_args['clear_nsec3'] = True
        return dnssec_args

    def _update_dnsssec_postprocess(self, db_session, update_info, dnssec_args):
        """
        Post update processing for checking DNSSEC state
        """
        if not update_info:
            # Things did not go as well as we thought...
            return
        if self.auto_dnssec:
            if not dnssec_args:
                # Empty dnssec_args flag no DNSSSEC keys present for zone.
                # This is only needed for enabling DNSSEC
                return
            if not update_info.get('dnskey_flag'):
                zone_sm_dnssec_schedule(db_session, self, 'sign')
                msg = ("Zone '%s' - DNSSEC configured and not DNSSEC signed"
                        % self.name)
                raise DynDNSUpdateRetry(msg)
            if self.nsec3:
                if not update_info.get('nsec3param_flag'):
                    msg = ("Zone '%s' - NSEC3 configured and not converted"
                            % self.name)
                    raise DynDNSUpdateRetry(msg)
            else:
                if update_info.get('nsec3param_flag'):
                    msg = ("Zone '%s' - NSEC3 not configured"
                            " and NSEC3 present" % self.name)
                    raise DynDNSUpdateRetry(msg)
        elif not self.auto_dnssec:
            if (update_info and update_info.get('dnskey_flag')):
                msg = ("Zone '%s' - DNSSEC signed and DNSSEC not configured"                            % self.name)
                raise DynDNSUpdateRetry(msg)


    def _update(self, event, clear_edit_lock=False, 
                        process_inc_updates=False):
        """
        Update Zone on name server, mainly from PUBLISHED state
        """
        db_session = event.db_session
        # Get new zi_id from zi_candidate_id
        zi_id = self.zi_candidate_id
        zi_type = sql_types['ZoneInstance']
        # Get zi, if not found, fail gracefully
        err_string = ''
        try:
            zi = db_session.query(zi_type)\
                    .filter(zi_type.id_ == zi_id).one()
        except NoResultFound as exc:
            err_string = str(exc)
        if err_string:
            raise DynDNSUpdateFailed(err_string)

        # Get SOA twiddling parameters, if any
        candidate_soa_serial = event.py_parameters.get('candidate_soa_serial')
        wrap_soa_serial = event.py_parameters.get('wrap_soa_serial')
        force_soa_serial_update = event.py_parameters.get(
                                                    'force_soa_serial_update')
        do_soa_serial_update = (candidate_soa_serial != None 
                                        or wrap_soa_serial 
                                        or force_soa_serial_update)

        # Add in any incremental updates here
        zi = self._do_incremental_updates(db_session, zi, process_inc_updates)

        # Update Apex Records
        zi.update_apex(db_session)

        # Update Zone TTLs
        zi.update_zone_ttls()
        
        # Preprocessing for DNSSEC goes here
        dnssec_args = self._update_dnssec_preprocess()
        
        # Run update engine
        (rcode, msg, soa_serial, update_info) = update_engine['dyndns']\
                        .update_zone(self.name, zi,
                                db_soa_serial=self.soa_serial,
                                candidate_soa_serial=candidate_soa_serial,
                                force_soa_serial_update=do_soa_serial_update,
                                wrap_serial_next_time=wrap_soa_serial,
                                **dnssec_args)

        # Handle auto reset of Zone SM if DNS server is not configured
        if rcode == RCODE_RESET:
            msg = "reconfiguring ZoneSM as server not configured" 
            log_info(msg)
            return self._retry_reconfig(event, randomize=True)

        # Post processing for DNSSEC goes here.
        # Empty dnssec_args flag no DNSSSEC keys present for zone.
        self._update_dnsssec_postprocess(db_session, update_info, dnssec_args)

        if rcode == RCODE_ERROR:
            raise DynDNSUpdateRetry(msg)
        elif rcode == RCODE_FATAL:
            raise DynDNSUpdateFailed(msg)

        # Update ZI in zone_sm.zi_id and self.soa_serial here.
        # Somehow because of SQL Alchemy, a raise does not revert
        # values in SA instrumented data....
        zi.ptime = db_time(db_session)
        # Don't have to update self.zi as this data is being committed and
        # is finished with as part of this event.  Only processed on event
        # queue.
        self.zi_id = zi.id_
        self.zi_candidate_id = self.zi_id
        self.soa_serial = soa_serial
        self.state = ZSTATE_PUBLISHED
        if clear_edit_lock:
            self._reset_edit_lock()
        return (rcode, msg)

    def _edit_update(self, event):
        """
        Update from a locked edit session, from UPDATE state
        """
        if  (event.py_parameters['edit_lock_token'] != self.edit_lock_token):
            raise ZoneEditLocked(self.name, self.edit_lock_token,
                    self.locked_by, self.locked_at)
        old_state = self.state
        # Rest of code is mostly the same as for _update() above
        try:
            return self._update(event, clear_edit_lock=True, 
                                process_inc_updates=True)
        except DynDNSUpdateFailed:
            # if failure, transition to published, releasing lock
            self.state = ZSTATE_PUBLISHED
            self._reset_edit_lock()
            return (RCODE_NOCHANGE, 
                    "Exiting %s for zone '%s' - update failed"
                    % (old_state, self.name))
    
    # State Table
    _sm_table  = {  ZSTATE_BATCH_CREATE_1: {
                            ZoneSMBatchConfig:   _batch_config,
                            ZoneSMDoReset: _do_reset,
                            ZoneSMDisable: _event_failure,
                            ZoneSMEnable: _already_enabled,
                            ZoneSMDelete: _nuke_start,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_BATCH_CREATE_2: {
                            ZoneSMReconfigCheck: _reconfig_check,
                            ZoneSMDoReset: _do_reset,
                            ZoneSMDisable: _event_failure,
                            ZoneSMEnable: _already_enabled,
                            ZoneSMDelete: _delete,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_INIT: {
                            ZoneSMDoConfig: _do_config,
                            ZoneSMDoBatchConfig: _do_batch_config,
                            ZoneSMDoReset: _do_reset,
                            ZoneSMDisable: _event_failure,
                            ZoneSMEnable: _already_enabled,
                            ZoneSMDelete: _nuke_start,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_CREATE: {
                            ZoneSMDoReset: _do_reset,
                            ZoneSMRemoveZoneFiles: _remove_zone_files,
                            ZoneSMEditSaved: _edit_exit,
                            ZoneSMEdit: _edit,
                            ZoneSMEditTimeout: _edit_timeout,
                            ZoneSMEditExit: _edit_exit,
                            ZoneSMEditLockTickle: _edit_lock_tickle,
                            ZoneSMDisable: _disable,
                            ZoneSMEnable: _already_enabled,
                            ZoneSMEditSavedNoLock: _other_edit_saved_no_lock,
                            ZoneSMEditSaved: _other_edit_saved,
                            ZoneSMDelete: _delete,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMConfig:   _config,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_UNCONFIG: {
                            ZoneSMDoReset: _do_reset,
                            ZoneSMEditSaved: _edit_exit,
                            ZoneSMEdit: _edit,
                            ZoneSMEditTimeout: _edit_timeout,
                            ZoneSMEditExit: _edit_exit,
                            ZoneSMEditLockTickle: _edit_lock_tickle,
                            ZoneSMDisable: _disable,
                            ZoneSMEnable: _already_enabled,
                            ZoneSMEditSavedNoLock: _other_edit_saved_no_lock,
                            ZoneSMEditSaved: _other_edit_saved,
                            ZoneSMDelete: _delete,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMReconfigUpdate: _reconfig_update,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_DISABLED: {
                            ZoneSMDoReset: _reset_disabled,
                            ZoneSMEditSaved: _edit_exit,
                            ZoneSMEdit: _edit,
                            ZoneSMEditTimeout: _edit_timeout,
                            ZoneSMEditExit: _edit_exit,
                            ZoneSMEditLockTickle: _edit_lock_tickle,
                            ZoneSMEnable: _enable,
                            ZoneSMDisable: _already_disabled,
                            ZoneSMEditSavedNoLock: _other_edit_saved_no_lock,
                            ZoneSMEditSaved: _other_edit_saved,
                            ZoneSMDelete: _delete,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMRemoveZoneFiles: _remove_zone_files,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_PUBLISHED: {
                            ZoneSMDoReset: _do_reset,
                            ZoneSMUpdate: _update,
                            ZoneSMEditSavedNoLock: _edit_saved_no_lock,
                            ZoneSMEditSaved: _edit_saved_to_update,
                            ZoneSMEdit: _edit,
                            ZoneSMEditTimeout: _edit_timeout,
                            ZoneSMEditExit: _edit_exit,
                            ZoneSMEditLockTickle: _edit_lock_tickle,
                            ZoneSMDisable: _disable,
                            ZoneSMEnable: _already_enabled,
                            ZoneSMDelete: _delete,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMDoReconfig: _do_reconfig,
                            ZoneSMDoRefresh: _do_refresh,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_UPDATE:  {   
                        # No EditExit here as this state will only fail
                        # with a significant event code failure
                            ZoneSMDoReset: _do_reset,
                            ZoneSMEditUpdate: _edit_update,
                            ZoneSMEditSaved: _edit_not_locked,
                            ZoneSMEdit: _edit_locked,
                            ZoneSMEditExit: _update_edit_exit,
                            ZoneSMDelete: _delete,
                            ZoneSMNukeStart: _nuke_start,
                            ZoneSMDoReconfig: _do_reconfig,
                            ZoneSMDoRefresh: _do_refresh,
                            ZoneSMDisable: _disable,
                            ZoneSMEnable: _already_enabled,
                            ZoneSMDoSgSwap: _do_sg_swap,
                            ZoneSMDoSetAltSg: _do_set_alt_sg,
                            },
                    ZSTATE_DELETED: {  
                            ZoneSMUndelete: _undelete,
                            ZoneSMRemoveZoneFiles: _remove_zone_files,
                            ZoneSMDoDestroy: _do_destroy,
                                        },
                }
    # RESET state is a copy of CREATE state
    _sm_table[ZSTATE_RESET] = _sm_table[ZSTATE_CREATE]
   
    def _check_dnssec_keys(self):
        key_glob = (settings['master_dnssec_key_dir'] + '/K' 
                        + self.name + '*')
        key_files = glob.glob(key_glob)
        if (len(key_files) <= 0):
            return False
        if not os.path.isfile(key_files[0]):
            return False
        return True

    def write_config(self, include_file, db_session, server_acls, 
            replica_sg=None):
        """
        Fill out master server configuration template 
        """
        if self.is_not_configured():
            # Disabled and deleted zones are NOT in master config file!
            return
        # Exceptions for this caught by an outer try:
        # self._template_names indirection used so that config file settings
        # will take effect.
        do_auto_dnssec = False
        if self.auto_dnssec:
            do_auto_dnssec = self._check_dnssec_keys()
            if not do_auto_dnssec:
                log_error("Zone '%s' - DNSSEC not configured due to no keys"
                            % self.name)
        template_dir = settings['master_template_dir']
        if do_auto_dnssec:
            template_name = (template_dir + '/'
                                + settings[self._template_names[1]])
        else:
            template_name = (template_dir + '/' 
                                + settings[self._template_names[0]])

        template = read_template(template_name)
        # Remove dot at end of zone name as this gives more
        # human literate filenames
        filler_name = self.name[:-1] if self.name.endswith('.') \
                                    else self.name
        sg_server_acls = '%s;' % server_acls[self.sg.name]['acl_name']
        if self.alt_sg:
            sg_server_acls += ' %s;' % server_acls[self.alt_sg.name]['acl_name']
        if replica_sg:
            sg_server_acls += ' %s;' % server_acls[replica_sg.name]['acl_name']
        # Include also-notify directive in string to be written
        # to stop any trouble due to blank also-notify statement, as
        # well as Admin confusion.....
        also_notify = ''
        for server_sm in self.sg.servers:
             # Skip server if it is actually this server
             if server_sm.is_this_server():
                 del server_sm
                 continue
             # include disabled server, as access can be shut off
             # in IPSEC and firewall!
             also_notify += ("%s; "% server_sm.address)
             del server_sm
        if self.alt_sg:
            for server_sm in self.alt_sg.servers:
                 # Skip server if it is actually this server
                 if server_sm.is_this_server():
                     del server_sm
                     continue
                 # include disabled server, as access can be shut off
                 # in IPSEC and firewall!
                 also_notify += ("%s; "% server_sm.address)
                 del server_sm
        if replica_sg:
            for server_sm in replica_sg.servers:
                 # Skip server if it is actually this server
                 if server_sm.is_this_server():
                     del server_sm
                     continue
                 # include disabled servers, as access can be shut off
                 # in IPSEC and firewall!
                 also_notify += ("%s; "% server_sm.address)
                 del server_sm
        if also_notify.endswith(' '):
            also_notify = also_notify[:-1]
        filler = { 'name': filler_name, 
                'master_dyndns_dir': settings['master_dyndns_dir'],
                'sg_server_acls': sg_server_acls,
                'also_notify': also_notify}
        section = template % filler
        include_file.write(section)
    

def get_default_zone_data(db_session):
    """
    Return default zone data from zone_cfg table
    """
    zone_sm_data = {}
    fields = ['use_apex_ns', 'auto_dnssec', 'edit_lock', 'nsec3', 
                'inc_updates']
    for field in fields:
        value = zone_cfg.get_row(db_session, field)
        if value in ('true', 'True', 'TRUE'):
            zone_sm_data[field] = True
        elif value in ('false', 'False', 'FALSE'):
            zone_sm_data[field] = False
        elif value is None:
            zone_sm_data[field] = False
        else:
            raise ZoneCfgItemValueError(field, value)
    return zone_sm_data

def new_zone(db_session, type_, sectag=None, sg_name=None, reference=None,
        **kwargs_init):
    """
    Create a new zone of type_, add it to the db_session, persist it,
    and return object.
    """
    zone_sm_data = get_default_zone_data(db_session)
    for arg in kwargs_init:
        if kwargs_init[arg] is None:
            kwargs_init[arg] = zone_sm_data.get(arg)
    # Check that SG exists
    if not sg_name:
        sg_name = zone_cfg.get_row_exc(db_session, 'default_sg')
    if not sg_name in list_all_sgs(db_session):
        raise NoSgFound(sg_name)
    zone_sm = type_(**kwargs_init)
    zone_sm.state = ZSTATE_INIT
    db_session.add(zone_sm)
    ZoneSecTag = sql_types['ZoneSecTag']
    if sectag and sectag != ZoneSecTag(settings['admin_sectag']):
        # Need a new sectag instance to go with this zone 
        if isinstance(sectag, str):
            new_sectag = ZoneSecTag(sectag)
        else:
            new_sectag = ZoneSecTag(sectag.sectag)
        zone_sm.add_sectag(db_session, new_sectag)
    sg = find_sg_byname(db_session, sg_name, raise_exc=True)
    zone_sm.set_sg(sg)
    # Add reference for zone
    if not reference:
        reference = zone_cfg.get_row_exc(db_session, 'default_ref')
    ref_obj = new_reference(db_session, reference, return_existing=True)
    ref_obj.set_zone(zone_sm)
    db_session.flush()
    return zone_sm

def del_zone(db_session, zone):
    """
    Delete the given zone
    """
    # Delete it from the DB
    db_session.delete(zone)
    db_session.commit()
    # Delete the object
    del(zone)

def exec_zonesm(zone_sm, sync_event_type, exception_type=ZoneSmFailure,
                **event_kwargs):
    """
    Execute a synchronous event of the zone state machine
    """
    if not issubclass(sync_event_type, SMSyncEvent):
        raise TypeError("'%s' is not a Synchonous Event." % sync_event_type)

    event = sync_event_type(sm_id=zone_sm.id_,
                                    zone_id=zone_sm.id_,
                                    name=zone_sm.name,
                                    **event_kwargs)
    results = event.execute()
    if results['state'] != ESTATE_SUCCESS:
        if isinstance(event, ZoneSMEdit):
            results['locked_by'] = zone_sm.locked_by
            results['locked_at'] = zone_sm.locked_at
        # By std Python convention exceptions don't have default value
        # arguments. Do the following to take care of 2 or 3 argument
        # variants for the exception.
        zi_id = event_kwargs.get('zi_id')
        if zi_id:
            raise exception_type(zone_sm.name, results['message'], results,
                                zi_id)
        else:
            raise exception_type(zone_sm.name, results['message'], results)
    return results

# SQL Alchemy hooks
def init_zone_sm_table():
    table = Table('sm_zone', sql_data['metadata'],
                        autoload=True, 
                        autoload_with=sql_data['db_engine'])
    sql_data['tables'][ZoneSM] = table

def init_zone_sm_mappers():
    table = sql_data['tables'][ZoneSM]
    sql_data['mappers'][ZoneSM] = mapper(ZoneSM, table,
            polymorphic_on=table.c.zone_type, 
            polymorphic_identity=ZoneSM.__name__,
            properties=mapper_properties(table, ZoneSM))
    sql_data['zone_sm_type_list'].append(ZoneSM.__name__)
    # Map all the zone_sm subclasses
    for class_ in sql_data['zone_sm_subclasses']:
        class_.sa_map_subclass()

sql_data['init_list'].append({'table': init_zone_sm_table,
                            'mapper': init_zone_sm_mappers})



#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Zone instance record incremental update code.  In an inherited
class to stop overpopulating zone_instance object.
"""


import copy

from magcode.core.globals_ import *
from magcode.core.database import sql_types
from dms.dns import *
from dms.database.zone_query import rr_query_db_raw
from dms.auto_ptr_util import check_auto_ptr_privilege


class ZiUpdate(object):
    """
    Container mix in class for ZI update code.
    """

    def __init__(self, db_session=None, trial_run=False, name=None):
        """
        This only for use with the PsuedoZi class, in zone_data_util.py
        """
        self.db_session = db_session
        self._trial_run = trial_run
        self.name = name
    
    def _rrop_find(self, query_rr, match_type=True, match_rdata=True):
        """
        Given an update_rr, find any matching records in this ZI

        Match is done using DNS python  and label rdata for accuracy.
        Type matches don't use dnspython data as dnspython rdata
        form may not exist.
        """
        # Some constants
        q_label = query_rr.dnspython_rr[0]
        q_type = query_rr.type_
        q_rdata = query_rr.dnspython_rr[2]
        # 1 Match label
        result = [rr for rr in self.rrs 
                if query_rr.dnspython_rr[0] == rr.dnspython_rr[0] ]
        # 2 Match type
        if not match_type or q_type == RRTYPE_ANY:
            return result
        result = [rr for rr in result if q_type == rr.type_]
        if not match_rdata or not q_rdata:
            return result
        # 3 Match RDATA
        result = [rr for rr in result if query_rr.dnspython_rr[2] == q_rdata]
        # return list of results
        return result

    def _rrop_finish(self, op_rr):
        """
        Detach a record from update group, and clear update_op
        """
        op_rr.unlink = True
        if hasattr(self, '_trial_run') and self._trial_run:
            return
        op_rr.update_op = None
        op_rr.ug_id = None
    
    def _rrop_update_rrtype(self, db_session, op_rr, sectag_label):
        """
        Implement update operation
        """
        # Find RRs
        old_rrs = self._rrop_find(op_rr, match_rdata=False)
        # update it
        for rr in old_rrs:
            self.remove_rr(rr)
        self.add_rr(op_rr)
        self._rrop_finish(op_rr)


    def _rrop_add(self, db_session, op_rr, sectag_label):
        """
        Implement add operation
        """
        # Find RRs
        old_rrs = self._rrop_find(op_rr)
        # Exit if it already exists!
        if len(old_rrs):
            self._rrop_finish(op_rr)
            return 
        self.add_rr(op_rr)
        self._rrop_finish(op_rr)

    def _rrop_delete(self, db_session, op_rr, sectag_label):
        """
        Implement delete operation
        """
        # Find RRs
        old_rrs = self._rrop_find(op_rr)
        # Delete RRs
        for rr in old_rrs:
            self.remove_rr(rr)
            if hasattr(self, '_trial_run') and self._trial_run:
                continue
            db_session.delete(rr)

    def _rrop_ptr_update(self, db_session, op_rr, sectag_label, force=False):
        """
        Implement PTR update operation on zone
        """
        # Handle auto reverse disbale settings
        if not settings['auto_reverse']:
            log_debug("Zone '%s' - can't process '%s' "
                                "- auto_reverse_enable off."
                        % (self.zone.name, op_rr.label))
            return
        # make proper sectag
        ZoneSecTag = sql_types['ZoneSecTag']
        sectag = ZoneSecTag(sectag_label)
        # Find old RRs
        old_rrs = self._rrop_find(op_rr, match_rdata=False)
        if len(old_rrs) > 1:
            log_error("Zone '%s' - multple PTR records for '%s', "
                        "contrary to RFC 1035 Section 3.5, not updating."
                        % (self.zone.name, op_rr.label))
            return
        old_rr = old_rrs[0] if len(old_rrs) else None
        
        # Check that we can proceed
        if not check_auto_ptr_privilege(op_rr.reference,
                sectag, self.zone, old_rr):
            if old_rr:
                log_debug("Zone '%s' - can't replace '%s' PTR as neither"
                    " sectags '%s' vs '%s'"
                    " references '%s' vs '%s'/'%s' (old PTR/rev zone) match,"
                    " or values not given."
                    % (self.zone.name, old_rr.label,
                        sectag.sectag, settings['admin_sectag'],    
                        op_rr.reference, old_rr.reference, self.zone.reference))
            else:
                log_debug("Zone '%s' - can't add '%s' PTR as neither"
                        " sectags '%s' vs '%s'"
                        " references '%s' vs '%s' (rev zone) match,"
                        " or values not given."
                    % (self.zone.name, op_rr.label,
                        sectag.sectag, settings['admin_sectag'],    
                        op_rr.reference, self.zone.reference))
            return
       
        if not force: 
            # Don't auto create PTRs if its not enabled.
            if (self.zone.name.endswith('in-addr.arpa.')
                    and not settings['auto_create_ipv4_ptr']):
                log_debug("Zone '%s' - can't process '%s' "
                                        "- auto_create_ipv4_ptr off."
                                % (self.zone.name, op_rr.label))
                return
            elif (self.zone.name.endswith('ip6.arpa.')
                    and not settings['auto_create_ipv6_ptr']):
                log_debug("Zone '%s' - can't process '%s' "
                                        "- auto_create_ipv6_ptr off."
                                % (self.zone.name, op_rr.label))
                return

        # See if forward still exists in DB
        if old_rr:
            if old_rr.lock_ptr:
                # Can't change if record locked
                log_debug("Zone '%s' - can't replace '%s' PTR as it is locked."
                    % (self.zone.name, old_rr.label)) 
                return
            # if new PTR same as old, ignore it!
            if op_rr == old_rr:
                log_debug("Zone '%s' - not replacing as '%s' PTR still the "
                        "same - rdata '%s'." 
                    % (self.zone.name, old_rr.label,
                        old_rr.rdata))
                return
            if not force:
                if self.zone.name.endswith('in-addr.arpa.'):
                    type_ = RRTYPE_A
                elif self.zone.name.endswith('ip6.arpa.'):
                    type_ = RRTYPE_AAAA
                else:
                    log_debug("Zone '%s' - can't determine Ip address "
                                            "record type" % self.zone.name)
                    return
                address = address_from_label(op_rr.label + '.' + self.zone.name)
                if not address:
                    log_warn("Zone '%s' - can't determine IP address from "
                            "label '%s' and domain!" 
                            % (self.zone.name, op_rr.label))
                    return
                result = rr_query_db_raw(db_session, 
                        label=old_rr.rdata, type_=type_, rdata=address)
                if result and len(result.get('rrs', [])):
                    # can't replace as old forward still active in DB
                    log_debug("Zone '%s' - can't replace '%s' PTR as old PTR "
                            "still valid - '%s'." 
                            % (self.zone.name, old_rr.label, 
                                old_rr.rdata))
                    return
            # Remove old RRs and update PTR
            for rr in old_rrs:
                self.remove_rr(rr)
        self.add_rr(op_rr)
        self._rrop_finish(op_rr)

    def _rrop_ptr_update_force(self, db_session, op_rr, sectag_label):
        """
        Implement PTR update operation on zone
        """
        self._rrop_ptr_update(db_session, op_rr, sectag_label, force=True)

    def trial_op_rr(self, op_rr):
        """
        Do a trial run of the operation
        """
        if not self._trial_run:
            # Should throw an Exception here.
            raise IncrementalUpdateNotInTrialRun(self.name)
        if not self._update_op_map.get(op_rr.update_op):
            log_error("Zone '%s': no method for update operation '%s'"
                        % (self.name, op_rr.update_op))
        op_rr.unlink = False
        # This is only trial execution of operation on zone that is 
        # already retrieved with sectag evaluation, sectag only for auto PTR
        # operations on a reverse zone
        sectag = None
        self._update_op_map[op_rr.update_op](self, self.db_session, 
                                                    op_rr, sectag)


    def exec_update_group(self, db_session, update_group):
        """
        Run an update group, and then clean up
        """
        ops_list = copy.copy(update_group.update_ops)
        for op_rr in ops_list:
            op_rr.unlink = False 
            if not self._update_op_map.get(op_rr.update_op):
                log_error("Zone '%s': no method for update operation '%s'"
                            % (self.zone.name, op_rr.update_op))
                continue
            self._update_op_map[op_rr.update_op](self, db_session, op_rr, 
                                                    update_group.sectag)
        # only go further than this if trial run
        if hasattr(self, '_trial_run') and self._trial_run:
            return
        # Complete addition of op_rrs to ZI
        for op_rr in ops_list:
            if not op_rr.unlink:
                db_session.delete(op_rr)
                continue
            update_group.update_ops.remove(op_rr)
        # Record source of the change
        if update_group.change_by:
            self.change_by = update_group.change_by
        # CASCADE constraint will remove all op_rrs not unlinked above
        db_session.delete(update_group)
        db_session.flush()

    # Map update operation functions to their OP tags   
    _update_op_map = {RROP_UPDATE_RRTYPE: _rrop_update_rrtype,
                        RROP_ADD: _rrop_add,
                        RROP_DELETE: _rrop_delete,
                        RROP_PTR_UPDATE: _rrop_ptr_update,
                        RROP_PTR_UPDATE_FORCE: _rrop_ptr_update_force
                        }




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#

import re

import dns
import dns.name
import dns.rdatatype
import dns.rdataclass
import dns.ttl
import dns.rdata
import dns.exception

from sqlalchemy.orm import reconstructor
from sqlalchemy.orm import relationship

from magcode.core.database import *
from dms.dns import *
from dms.exceptions import *
import dms.database.rr_comment


# Dictionary for mapping Record Type to Resource Record Class
rrtype_map = {}

# Lists we use in global sql_data dict
sql_data['rr_subclasses'] = []
sql_data['rr_type_list'] = []


def rr_register(class_):
    """
    Resorce record descedant class decorator function to register class for SQL
    Alchemy mapping in init_rr_class() below, called from
    magcode.core.database.utility.setup_sqlalchemy()
    """
    sql_data['rr_subclasses'].append(class_)
    rrtype_map[class_._rr_type] = class_
    # Also add as an SQL data type
    typeregister(class_)
    return(class_)

@typeregister
class ResourceRecord(object):
    """
    DNS Resource Record type.
    """
   
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_NULL

    @classmethod
    def sa_map_subclass(class_):
        metadata = MetaData()
        sql_data['mappers'][class_] = mapper(class_,
                    inherits=sql_data['mappers'][ResourceRecord], 
                    polymorphic_identity=class_._rr_type)
        sql_data['rr_type_list'].append(class_._rr_type)
    
    @classmethod
    def _mapper_properties(class_):
        rr_table = sql_data['tables'][sql_types['ResourceRecord']]
        rr_comment_table = sql_data['tables'][sql_types['RRComment']]
        ref_type = sql_types['Reference']
        ref_table = sql_data['tables'][ref_type]
        return {'rr_comment': relationship(sql_types['RRComment'],
                                    primaryjoin=(rr_table.c.comment_rr_id 
                                            == rr_comment_table.c.get('id')),
                                        uselist=False,
                                        backref="rr"),
                'group_comment': relationship(sql_types['RRComment'],
                                    primaryjoin=(rr_table.c.comment_group_id 
                                            == rr_comment_table.c.get('id')),
                                        backref="rr_group"),
                'reference': relationship(sql_types['Reference'],
                                        viewonly=True),
                }

    def __init__(self, label=None, ttl=None, zone_ttl=None, rdata=None,
            domain=None, dnspython_rr=None, comment_rr_id=None, 
            comment_group_id=None, lock_ptr=False, disable=False, ref_id=None,
            ug_id=None, update_op=None, track_reverse=False, **kwargs):
        """
        Initialise resource record, and build private
        dnspython rdata tuple for comparison  and parse checking purposes.
        """
        # Swallow type and class arguments, otherwise raise exception
        kwargs = [arg for arg in kwargs if (arg != 'type' and arg != 'class')]
        if kwargs:
            raise TypeError(
                    "__init__() got an unexpected keyword argument '%s'" 
                    % kwargs[0])

        if domain:
            # Check that domain ends in '.'
            if not domain.endswith('.'):
                raise dns.exception.SyntaxError(
                                    "domain '%s' must end with '.'." 
                                    % domain)
            # Relativize label
            label = relativize_domain_name(label, domain)
            # label should not now be an FQDN!
            if label.endswith('.'):
                raise dns.exception.SyntaxError(
                                "FQDN label '%s' is not within domain '%s'."
                                % (label, domain))
            # Make domain dnspython dns.name.Name
            domain = dns.name.from_text(domain)

        # Check that rdata IS supplied
        # This closes hole in error handling - these are the only
        # times when rdata can be blank
        if (not rdata and not dnspython_rr and self._rr_type != RRTYPE_ANY and 
                update_op != RROP_DELETE):
            raise ValueError("rdata must be supplied")

        self.label = label
        self.ttl = ttl
        self.zone_ttl = zone_ttl
        self.rdata = rdata
        self.comment_rr_id = comment_rr_id
        self.comment_group_id = comment_group_id
        self.lock_ptr = lock_ptr
        self.disable = disable
        self.ref_id = ref_id
        self.update_op = update_op
        self.ug_id = ug_id
        self.track_reverse = track_reverse
        self.type_ = self._rr_type
        self.class_ = self._rr_class
        if dnspython_rr:
            # if given dnspython_rr tuple, assume it is already
            # relativized by dnspython (default in zone scanning/parsing)
            self.dnspython_rr = dnspython_rr
            self.label = str(self.dnspython_rr[0])
            self.ttl = str(self.dnspython_rr[1])
            self.rdata = str(self.dnspython_rr[2])
        else:
            self._dnspython_from_rdata(domain)
            # relativize rdata according to domain via dnspython
            if self.dnspython_rr[2]:
                self.rdata = str(self.dnspython_rr[2])
    
    def _get_dnspython_ttl(self):
        if (self.ttl is not None):
            ttl = dns.ttl.from_text(self.ttl)
        elif (self.zone_ttl is not None):
            ttl = dns.ttl.from_text(self.zone_ttl)
        else:
            if self.id_:
                raise ZoneTTLNotSetError(self.id_)
            else:
                raise ValueError("RR zone_ttl can not be None")
        return ttl

    def _dnspython_from_rdata(self, domain=None):
        """
        Use the dnspython from_text() methods and its tokenizer to initialise
        dnspython rdata.
        """
        if (domain):
            origin = domain
        else:
            origin = dns.name.empty
        label = dns.name.from_text(self.label, origin)
        label = label.choose_relativity(origin)
        rdtype = dns.rdatatype.from_text(self.type_)
        rdclass = dns.rdataclass.from_text(self.class_)
        ttl = self._get_dnspython_ttl()
        # Bob Halley dnspython author recommended the following as it is an
        # API call rather than a dig into the guts of dnspython.
        if self.rdata:
            rdata = dns.rdata.from_text(rdclass, rdtype, self.rdata, origin)
        else:
            rdata = None
        self.dnspython_rr = [label, ttl, rdata]

    @reconstructor
    def rr_reconstructor(self):
        """
        Reconstruct dnspython rdata and rdata_dict from rdata,
        when loading from SQLAlchemy
        """
        # Close hole in error handling - these are the only
        # times when rdata can be blank
        if (not self.rdata and self.type_ != RRTYPE_ANY and 
                self.update_op != RROP_DELETE):
            raise ValueError("RR(%s) - rdata must not be blank" % self.id_)
        # Complete initialisation from sqlalchemy
        self._dnspython_from_rdata()

    def __eq__(self, other):
        """
        Compare rdata records for equality
        """
        return self.dnspython_rr == other.dnspython_rr

    def __ne__(self, other):
        """
        Compare rdata records for inequality
        """
        return self.dnspython_rr != other.dnspython_rr

    def __lt__(self, other):
        """
        Compare rdata records for inequality
        """
        return self.dnspython_rr < other.dnspython_rr

    def __le__(self, other):
        """
        Compare rdata records for inequality
        """
        return self.dnspython_rr <= other.dnspython_rr

    def __gt__(self, other):
        """
        Compare rdata records for inequality
        """
        return self.dnspython_rr > other.dnspython_rr

    def __ge__(self, other):
        """
        Compare rdata records for inequality
        """
        return self.dnspython_rr >= other.dnspython_rr

    def _rr_str(self):
        """
        Common code between __repr__ and __str__
        """
        if self.dnspython_rr[2]:
            rdata = self.dnspython_rr[2].__str__()
        elif self.rdata:
            rdata = self.rdata
        else:
            rdata = None
        stuff = [self.label, self.class_, self.type_]
        if rdata:
            stuff.append(rdata)
        string =  ' '.join(stuff)
        return string

    def __str__(self):
        """
        String representation of rdata
        """
        return self._rr_str()

    def __repr__(self):
        """
        Mnemonic representation of rdata
        """
        string = self._rr_str()
        return '<'+ self.__class__.__name__ + " '" + string + "'>"

    def to_engine_brief(self, time_format=None):
        """
        Output for zone engine.
        """
        reference = self.reference.reference \
                if hasattr(self, 'reference') and self.reference else None
        return{'rr_id': self.id_, 'zi_id': self.zi_id,
                'label': self.label,
                'ttl': self.ttl,
                'class': self.class_, 'type': self.type_, 'rdata': self.rdata,
                'comment_group_id': self.comment_group_id, 
                'comment_rr_id': self.comment_rr_id,
                'lock_ptr': self.lock_ptr,
                'disable': self.disable,
                'track_reverse': self.track_reverse,
                'reference': reference}

    to_engine = to_engine_brief

    def _update_dnspython_ttl(self):
            if self.ttl:
                self.dnspython_rr[1] = dns.ttl.from_text(self.ttl)
            else:
                self.dnspython_rr[1] = dns.ttl.from_text(self.zone_ttl)

    def update_zone_ttl(self, zone_ttl, reset_rr_ttl=False):
        """
        Update default ttl, and if ttl = self.ttl
        set in to None
        """ 
        if type(zone_ttl) == str:
            z_ttl = dns.ttl.from_text(zone_ttl)
        else:
            z_ttl = zone_ttl
        if (reset_rr_ttl and self.ttl):
            rr_ttl = dns.ttl.from_text(self.ttl) 
            if (rr_ttl == z_ttl):
                self.ttl = None
        self.zone_ttl = str(zone_ttl)
        self._update_dnspython_ttl()

    def update_ttl(self, ttl):
        """
        Update ttl, converting from integer if required
        """
        self.ttl = str(ttl)
        self._update_dnspython_ttl

    def get_effective_ttl(self):
        """
        returns effective ttl as a string
        """
        if self.ttl:
            return self.ttl
        else:
            return self.zone_ttl

# RR_SOA is most complex RR class to to need to manipulate rdata,
# ONLY RR this is done for - makes SQLAlchemy database persistance easier.
@rr_register
class RR_SOA(ResourceRecord):
    """
    SOA record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_SOA

    def update_serial(self, serial):
        self.dnspython_rr[2].serial = serial
        self.rdata = str(self.dnspython_rr[2])

    def get_serial(self):
        serial = self.dnspython_rr[2].serial
        return serial

@rr_register
class RR_CNAME(ResourceRecord):
    """
    CNAME record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_CNAME

@rr_register
class RR_NS(ResourceRecord):
    """
    NS record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_NS

@rr_register
class RR_A(ResourceRecord):
    """
    A record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_A

@rr_register
class RR_AAAA(ResourceRecord):
    """
    A record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_AAAA

@rr_register
class RR_PTR(ResourceRecord):
    """
    NS record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_PTR

@rr_register
class RR_TXT(ResourceRecord):
    """
    TXT record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_TXT

@rr_register
class RR_MX(ResourceRecord):
    """
    MX record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_MX

@rr_register
class RR_SPF(ResourceRecord):
    """
    SPF record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_SPF

@rr_register
class RR_RP(ResourceRecord):
    """
    RP record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_RP

@rr_register
class RR_SSHFP(ResourceRecord):
    """
    SSHFP record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_SSHFP

@rr_register
class RR_SRV(ResourceRecord):
    """
    SRV record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_SRV

@rr_register
class RR_NSAP(ResourceRecord):
    """
    SRV record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_NSAP

@rr_register
class RR_NAPTR(ResourceRecord):
    """
    SRV record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_NAPTR

@rr_register
class RR_LOC(ResourceRecord):
    """
    LOC record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_LOC

@rr_register
class RR_KX(ResourceRecord):
    """
    KX record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_KX

@rr_register
class RR_IPSECKEY(ResourceRecord):
    """
    IPSECKEY record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_IPSECKEY

@rr_register
class RR_HINFO(ResourceRecord):
    """
    HINFO record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_HINFO

@rr_register
class RR_CERT(ResourceRecord):
    """
    CERT record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_CERT

@rr_register
class RR_DS(ResourceRecord):
    """
    DS record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_DS

@rr_register
class RR_ANY(ResourceRecord):
    """
    NULL record type for use in 
    RR_DELETE operations
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_ANY
    
@rr_register
class RR_TLSA(ResourceRecord):
    """
    TLSA DANE record type
    """
    _rr_class = RRCLASS_IN
    _rr_type = RRTYPE_TLSA

# Factory functions
def dnspython_to_rr(dnspython_rr):
    """
    Factory Function to set up an RR from dnspython data
    """
    type_ = dns.rdatatype.to_text(dnspython_rr[2].rdtype)
    class_ = rrtype_map[type_]
    return class_(dnspython_rr=dnspython_rr)


def _lower_case_names(domain, rr_data):
    """
    lower case labels, and hostnames in PTR, NS, SOA, MX and SRV records
    to prevent DNSSEC duplicates.
    """
    # Lower case label
    rr_data['label'] = rr_data['label'].lower()
    # Deal with SOA NS MX and PTR records lower casing a digit gives the digit!
    if rr_data['type'] in (RRTYPE_MX, RRTYPE_NS, RRTYPE_SOA, RRTYPE_SRV):
        rr_data['rdata'] = rr_data['rdata'].lower()

def _check_name(domain, rr, rr_data):
    """
    Check host names in RRs that bind checks to conform hostname specifications

    Onwer labels on A, AAAA, and MX records
    Domain names in RDATA of SOA, NS, MX, SRV, and PTR records in IN-ADDR.ARPA,
    IP6.ARPA, or IP6.INT (Bind Option check-names)
    """
    if rr.type_ in (RRTYPE_MX, RRTYPE_A, RRTYPE_AAAA):
        if not is_inet_hostname(rr.label):
            raise BadNameOwnerError(domain, rr_data)
    if rr.type_ == RRTYPE_NS:
        if not is_inet_hostname(rr.rdata):
            raise BadNameRdataError(domain, rr_data, rr.rdata)
    if (rr.type_ == RRTYPE_PTR
                and (domain.upper().endswith('IN-ADDR.ARPA.')
                        or domain.upper().endswith('IP6.ARPA.')
                            or domain.upper().endswith('IP6.INT.'))):
        if not is_inet_hostname(rr.rdata, wildcard=False):
            raise BadNameRdataError(domain, rr_data, rr.rdata)
    if (rr.type_ == RRTYPE_MX):
        rdata = rr.rdata.split()
        if len(rdata) != 2:
            raise RdataParseError(domain, rr_data, 'MX record has 2 fields')
        if not  is_inet_hostname(rdata[1]):
            raise BadNameRdataError(domain, rr_data, rdata[1])
    if (rr.type_ == RRTYPE_SRV):
        rdata = rr.rdata.split()
        if len(rdata) != 4:
            raise RdataParseError(domain, rr_data, 'SRV record has 4 fields')
        if not  is_inet_hostname(rdata[3]):
            raise BadNameRdataError(domain, rr_data, rdata[3])
    if (rr.type_ == RRTYPE_SOA):
        rdata = rr.rdata.split()
        if len(rdata) != 7:
            raise RdataParseError(domain, rr_data, 'SOA record has 7 fields')
        if not  is_inet_hostname(rdata[0]):
            raise BadNameRdataError(domain, rr_data, rdata[0])
        if not  is_inet_hostname(rdata[1]):
            raise BadNameRdataError(domain, rr_data, rdata[1])

def data_to_rr(domain, rr_data):
    """
    Factory Function that returns resource record based on
    incoming data

    This is big because of the need for trapping errors in dnspython, 
    and producing useful error feedback
    """
    class_ = rr_data.get('class')
    if class_:
        class_ = rr_data['class'] = rr_data['class'].upper()
        if (class_ != RRCLASS_IN):
            raise UnhandledClassError(domain, rr_data)
    type_ = rr_data.get('type')
    if not type_:
        raise RRNoTypeGiven(domain, rr_data)
    type_ = rr_data['type'] = rr_data['type'].upper()
    if type_ not in rrtype_map.keys():
        raise UnhandledTypeError(domain, rr_data)

    # ANY type only allowed with RROP_DELETE
    if rr_data.get('update_op') != RROP_DELETE and type_ == RRTYPE_ANY:
        raise UnhandledTypeError(domain, rr_data)

    # update_op must be a valid value
    update_op = rr_data.get('update_op')
    if update_op not in rr_op_values:
        raise InvalidUpdateOperation(domain, rr_data)

    class_ = rrtype_map[type_]

    # Check that label is going to be useful!
    # Relativize label
    _lower_case_names(domain, rr_data)
    label = relativize_domain_name(rr_data['label'], domain)
    # label should not now be an FQDN!
    if label.endswith('.'):
        raise LabelNotInDomain(domain, rr_data)

    # Process recieved data
    try:
        kwargs = rr_data.copy()
        kwargs.pop('rdata_pyparsing', None)
        kwargs.pop('rr_groups_index', None)
        kwargs.pop('rrs_index', None)
        kwargs.pop('force_reverse', None)
        kwargs.pop('reference', None)
        rr = class_(domain=domain, **kwargs)
    except Exception as exc:
        err_string = str(exc)
        raise RdataParseError(domain, rr_data, msg=err_string)

    # Do the Bind9 bad names check
    _check_name(domain, rr, rr_data)
    return  rr

def relativize_domain_name(domain_name, zone_name):
    """
    Relativizes domain name wrt to a zone name
    """
    if domain_name.endswith(zone_name):
        d_index = domain_name.rfind(zone_name)
        if d_index == 0:
            domain_name = '@'
        elif d_index:
            domain_name = domain_name[:d_index-1]
    return domain_name

# SQL Alchemy hooks
def init_rr_table():
    table = Table('resource_records', sql_data['metadata'],
                        autoload=True, 
                        autoload_with=sql_data['db_engine'])
    sql_data['tables'][ResourceRecord] = table

def init_rr_mappers():
    table = sql_data['tables'][ResourceRecord]
    sql_data['mappers'][ResourceRecord] = mapper(ResourceRecord, table,
            polymorphic_on=table.c.get('type'), 
            polymorphic_identity=ResourceRecord._rr_type,
            properties=mapper_properties(table, ResourceRecord))
    sql_data['rr_type_list'].append(ResourceRecord._rr_type)
    # Map all the event subclasses
    for class_ in sql_data['rr_subclasses']:
        class_.sa_map_subclass()

sql_data['init_list'].append({'table': init_rr_table, 'mapper': init_rr_mappers})




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Server Groups
"""


import io
import errno
import tempfile
from os.path import basename

from sqlalchemy.orm import relationship
from sqlalchemy.orm import backref

from magcode.core.utility import get_numeric_setting
from magcode.core.database import *
import dms.database.server_sm
from dms.template_cache import read_template


@saregister
class ServerGroup(object):
    """
    Server Group 
    """
    _table = 'server_groups'
    
    @classmethod
    def _mapper_properties(class_):
        zone_sm_type = sql_types['ZoneSM']
        server_sm_type = sql_types['ServerSM']
        zone_cfg_type = sql_types['ZoneCfg']
        sg_table = sql_data['tables'][ServerGroup]
        zone_sm_table = sql_data['tables'][zone_sm_type]
        return {'zones': relationship(zone_sm_type, 
                            primaryjoin=(zone_sm_table.c.sg_id 
                                        == sg_table.c.get('id')),
                                    passive_deletes=True, lazy='dynamic'),
                'alt_zones': relationship(zone_sm_type, 
                            primaryjoin=(zone_sm_table.c.alt_sg_id 
                                        == sg_table.c.get('id')),
                                    passive_deletes=True, lazy='dynamic'),
                'servers': relationship(server_sm_type, passive_deletes=True,
                                    backref='sg'),
                'zone_cfg_entries': relationship(zone_cfg_type,
                                    passive_deletes=True,
                                    backref='sg'),
                            }

    def __init__(self, sg_name, config_dir, master_address, 
            master_alt_address):
        """
        Create an SG object
        """
        self.name = sg_name
        self.config_dir = config_dir
        self.master_address = master_address
        self.master_alt_address = master_alt_address

    def __eq__(self, other):
        if not other:
            return False
        return self.name == other.name

    def __ne__(self, other):
        if not other:
            return True
        return self.name != other.name

    def to_engine_brief(self, time_format=None):
        """
        Output server group attributes as JSON
        """
        config_dir = self.config_dir if self.config_dir \
                        else settings['server_config_dir']
        replica_sg = (True if hasattr(self, 'master_sm') 
                            and self.master_sm else False)
        return {'sg_id': self.id_, 'sg_name': self.name,
                'config_dir': config_dir,
                'master_address': self.master_address,
                'master_alt_address': self.master_alt_address,
                'replica_sg': replica_sg,
                'zone_count': self.zone_count}

    # Use assignment to fill out to_engine() method
    to_engine = to_engine_brief

    def get_include_dir(self):
        """
        Function to return include dir for SG
        """
        include_dir = settings['sg_config_dir'] + '/' + self.name
        return include_dir

    def get_include_file(self, server_type):
        """
        Function to return the include file path for a server type.
        """
        include_dir = settings['sg_config_dir'] + '/' + self.name
        include_file = include_dir + '/' + server_type + '.conf'
        return include_file

    def write_config(self, db_session, op_exc):
        """
        Write out all configuration files needed for a server group.
        """
        def write_zone_include(zone_sm):
            # Remove dot at end of zone name as this gives more
            # human literate filenames
            filler_name = zone_sm.name[:-1]  \
                    if zone_sm.name.endswith('.') \
                            else zone_sm.name

            filler = {'name': filler_name, 
                    'master': master_address}
            tmp_file.write(template % filler)
        
        replica_sg = (True if hasattr(self, 'master_sm') 
                            and self.master_sm else False)
        # Calculate master addresses
        if (self.master_address 
                    and self.master_address 
                    in settings['this_servers_addresses']):
                master_address = self.master_address
        elif (self.master_alt_address
                    and self.master_alt_address 
                    in settings['this_servers_addresses']):
                master_address = self.master_alt_address
        else:
            master_address = settings['master_dns_server']

        # Get list of server types in SG
        ServerSM = sql_types['ServerSM']       
        server_types = [s.server_type for s in self.servers]
        # sort|uniq the types list
        server_types = list(set(sorted(server_types)))
        if replica_sg:
            server_types = [ st + settings['server_replica_suffix'] 
                            for st in server_types ]
        db_query_slice = get_numeric_setting('db_query_slice', int)
        for server_type in server_types:
            include_dir = self.get_include_dir()
            include_file = self.get_include_file(server_type)
            if self.config_dir:
                # This allows us to override default template configuration
                # for say internal domains which IPV6 ULA/
                # IPV4 RFC1918 addressing
                template_file = (self.config_dir  + '/'
                                    + server_type + '.conf')
            else:
                template_file = (settings['server_config_dir'] + '/'
                                    + server_type + '.conf')
            try:
                # Make directory if it already does not exist
                # This is in here to avoid try: verbosity
                if not os.path.isdir(include_dir):
                    os.mkdir(include_dir)
                template = read_template(template_file)
                (fd, tmp_filename) = tempfile.mkstemp(
                            dir=include_dir,
                            prefix='.' + basename(include_file) + '-')
                tmp_file = io.open(fd, mode='wt')
                zone_sm_type = sql_types['ZoneSM']
                zone_count = 0
                if replica_sg:
                    # Master SG  - include all zones
                    query = db_session.query(zone_sm_type)
                    query = zone_sm_type.query_sg_is_configured(query)\
                                .yield_per(db_query_slice)
                    for zone_sm in query:
                        write_zone_include(zone_sm)
                        zone_count += 1
                        # Prevent Memory leaks...
                        del zone_sm
                else:
                    query = zone_sm_type.query_sg_is_configured(self.zones)\
                                .yield_per(db_query_slice)
                    for zone_sm in query:
                        write_zone_include(zone_sm)
                        zone_count += 1
                        # Prevent Memory leaks...
                        del zone_sm
                    query = zone_sm_type.query_sg_is_configured(
                                self.alt_zones)\
                                .yield_per(db_query_slice)
                    for zone_sm in query:
                        write_zone_include(zone_sm)
                        zone_count += 1
                        # Prevent Memory leaks...
                        del zone_sm
                tmp_file.close()
                # Rename tmp file into place so that replacement is atomic
                os.chmod(tmp_filename, int(settings['config_file_mode'],8))
                os.rename(tmp_filename, include_file)
                # Store zone_count for monitoring data input 
                self.zone_count = zone_count

            except (IOError, OSError) as exc:
                msg = ( "SG %s - '%s' - %s." 
                                % (self.name, exc.filename, exc.strerror))
                if exc.errno in (errno.ENOENT, errno.EPERM, errno.EACCES):
                    raise op_exc(msg)
                else:
                    raise exc
            except KeyError as exc:
                msg = ("SG %s - Invalid template key in template file %s - %s"
                        % (self.name, template_file, str(exc)))
                raise op_exc(msg)
            finally:
                # clean up if possible
                try:
                    os.unlink(tmp_filename)
                except:
                    pass
        return






#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""Test Utility

Program implemented by subclassing net.24.core.process.Process, and 
replacing the main() method.
"""


import os
import sys
import io
import re
import tempfile
import stat
import socket
from subprocess import check_call
from subprocess import CalledProcessError

import dns.zone

from magcode.core.process import Process
from magcode.core.process import BooleanCmdLineArg
from magcode.core.process import BaseCmdLineArg
from magcode.core.globals_ import *
from dms.dyndns_update import DynDNSUpdate
from dms.database.resource_record import dnspython_to_rr
from dms.database.zone_instance import ZoneInstance
from dms.exceptions import DynDNSCantReadKeyError
from dms.exceptions import NoSuchZoneOnServerError
from magcode.core.database import RCODE_OK
from magcode.core.database import RCODE_ERROR
from magcode.core.database import RCODE_RESET
from magcode.core.database import RCODE_FATAL
from magcode.core.database import RCODE_NOCHANGE


USAGE_MESSAGE = "Usage: %s [-dfhknprsuvy] [-c config_file] <domain-name> [dns-server]"
COMMAND_DESCRIPTION = "Edit or manipulate a domain directly via dynamic DNS"

settings['config_section'] = 'DEFAULT'

class NoSOASerialUpdateCmdLineArg(BooleanCmdLineArg):
    """
    Process verbose command Line setting
    """
    def __init__(self):
        BooleanCmdLineArg.__init__(self,
                                short_arg='n',
                                long_arg='no-serial',
                                help_text="Don't update SOA serial no",
                                settings_key = 'no_serial',
                                settings_default_value = False,
                                settings_set_value = True)

class WrapSOASerialCmdLineArg(BooleanCmdLineArg):
    """
    Process verbose command Line setting
    """
    def __init__(self):
        BooleanCmdLineArg.__init__(self,
                                short_arg='r',
                                long_arg='wrap-serial',
                                help_text="Wrap SOA serial no",
                                settings_key = 'wrap_serial',
                                settings_default_value = False,
                                settings_set_value = True)

class UpdateSOASerialCmdLineArg(BooleanCmdLineArg):
    """
    Process verbose command Line setting
    """
    def __init__(self):
        BooleanCmdLineArg.__init__(self,
                                short_arg='u',
                                long_arg='update-serial',
                                help_text="Just update SOA serial normally",
                                settings_key = 'update_serial',
                                settings_default_value = False,
                                settings_set_value = True)

class ForceUpdateCmdLineArg(BooleanCmdLineArg):
    """
    Process verbose command Line setting
    """
    def __init__(self):
        BooleanCmdLineArg.__init__(self,
                                short_arg='f',
                                long_arg='force-update',
                                help_text="Force update if file unchanged",
                                settings_key = 'force_update',
                                settings_default_value = False,
                                settings_set_value = True)

class ClearDnskeyCmdLineArg(BooleanCmdLineArg):
    """
    Process verbose command Line setting
    """
    def __init__(self):
        BooleanCmdLineArg.__init__(self,
                                short_arg='y',
                                long_arg='clear-dnskey',
                                help_text="Delete apex DNSKEY RRs",
                                settings_key = 'clear_dnskey',
                                settings_default_value = False,
                                settings_set_value = True)

class ClearNsec3CmdLineArg(BooleanCmdLineArg):
    """
    Process verbose command Line setting
    """
    def __init__(self):
        BooleanCmdLineArg.__init__(self,
                                short_arg='k',
                                long_arg='clear-nsec3',
                                help_text="Delete NSEC3PARAM RR",
                                settings_key = 'clear_nsec3',
                                settings_default_value = False,
                                settings_set_value = True)

class Nsec3SeedCmdLineArg(BooleanCmdLineArg):
    """
    Process verbose command Line setting
    """
    def __init__(self):
        BooleanCmdLineArg.__init__(self,
                                short_arg='s',
                                long_arg='nsec3-seed',
                                help_text="Create NSEC3PARAM RR",
                                settings_key = 'nsec3_seed',
                                settings_default_value = False,
                                settings_set_value = True)

class PortCmdLineArg(BaseCmdLineArg):
    """
    Handle configuration file setting
    """
    def __init__(self):
        BaseCmdLineArg.__init__(self, short_arg='p:',
                                long_arg="port=",
                                help_text="set DNS server port")
    
    def process_arg(self, process, value, *args, **kwargs):
        """
        Set configuration file name
        """
        settings['dns_port'] = value 


class DynDNSTool(Process):
    """
    Process Main Daemon class
    """
    def __init__(self, *args, **kwargs):
        Process.__init__(self, usage_message=USAGE_MESSAGE,
            command_description=COMMAND_DESCRIPTION, *args, **kwargs)
        self.cmdline_arg_list.append(NoSOASerialUpdateCmdLineArg())
        self.cmdline_arg_list.append(WrapSOASerialCmdLineArg())
        self.cmdline_arg_list.append(UpdateSOASerialCmdLineArg())
        self.cmdline_arg_list.append(ForceUpdateCmdLineArg())
        self.cmdline_arg_list.append(Nsec3SeedCmdLineArg())
        self.cmdline_arg_list.append(ClearNsec3CmdLineArg())
        self.cmdline_arg_list.append(ClearDnskeyCmdLineArg())
        self.cmdline_arg_list.append(PortCmdLineArg())

    def parse_argv_left(self, argv_left):
        """
        Handle any arguments left after processing all switches

        Override in application if needed.
        """
        if (len(argv_left) != 1 and len(argv_left) != 2):
            self.usage_short()
            sys.exit(os.EX_USAGE)
        
        self.argv_left = argv_left
        self.zone_name = argv_left[0]
        if not re.match('^[\S\.]+$', self.zone_name):
            self.usage_short()
            sys.exit(os.EX_USAGE)
        if (not self.zone_name.endswith('.')):
            self.zone_name += '.'

        if (len(argv_left) ==2):
            if not re.match('^[\S\.]+$', argv_left[1]):
                self.usage_short()
                sys.exit(os.EX_USAGE)
            settings['dns_server'] = argv_left[1]


    def _get_editor(self):
        """
        Work out the users preference of editor, and return that
        """
        editor = os.getenvb(b'VISUAL')
        if (editor):
            return editor
        
        editor = os.getenvb(b'EDITOR')
        if (editor):
            return editor

        editor = b'/usr/bin/sensible-editor'
        if os.path.isfile(editor):
            return editor 
        
        editor = b'/usr/bin/editor'
        if os.path.isfile(editor):
            return editor 

        # Fall back if none of the above is around...
        return b'/usr/bin/vi'

    def main_process(self):
        """Main process editzone
        """
        def clean_up():
            if (tmp_file):
                os.unlink(tmp_file)

        tmp_file = ''
        # Get update session object
        error_str = ''
        try:
            update_session = DynDNSUpdate(settings['dns_server'],
                                    settings['dyndns_key_file'],
                                    settings['dyndns_key_name'],
                                    )
        except (socket.error, DynDNSCantReadKeyError, IOError) as exc:
            error_str = str(exc)
        # Process above error...
        if (error_str):
            log_error("%s" % error_str)
            sys.exit(os.EX_NOHOST)

        # Do AXFR to obtain current zone data
        msg = None
        try:
            (zone, dnskey_flag, nesc3param_flag) \
                = update_session.read_zone(self.zone_name)
        except NoSuchZoneOnServerError as exc:
            msg = str(exc)
        if msg:
            log_error(msg)
            sys.exit(os.EX_NOINPUT)
            
        # Only edit zone if not wrapping SOA serial number 
        if (not settings['wrap_serial'] and not settings['update_serial']
                and not settings['nsec3_seed'] and not settings['clear_nsec3']
                and not settings['clear_dnskey']):
            # Write zone out to a temporary file
            (fd, tmp_file) = tempfile.mkstemp(prefix=settings['process_name'] + '-',
                                    suffix='.zone')
            os.close(fd)
            zone.to_file(tmp_file)

            # Edit zone data
            old_stat = os.stat(tmp_file)
            editor = self._get_editor()
            try:
                output = check_call([editor, tmp_file])
            except CalledProcessError as exc:
                log_error("editor exited with '%s'." % exc.returncode)
                sys.exit(os.EX_SOFTWARE)
                
            new_stat = os.stat(tmp_file)
            if (not settings['force_update'] 
                    and old_stat[stat.ST_MTIME] == new_stat[stat.ST_MTIME]
                    and old_stat[stat.ST_SIZE] == new_stat[stat.ST_SIZE]
                    and old_stat[stat.ST_INO] == new_stat[stat.ST_INO]):
                log_info("File '%s' unchanged after editing - exiting." % tmp_file)
                clean_up()
                sys.exit(os.EX_OK)
     
            # Read in file and form zi structure
            zone = dns.zone.from_file(tmp_file, self.zone_name)
        # At the moment these values are just for the sake of it.
        zi = ZoneInstance(soa_refresh='5m', soa_retry='5m', soa_expire='7d', soa_minimum='600')
        for rdata in zone.iterate_rdatas():
            zi.add_rr(dnspython_to_rr(rdata))

        # Update Zone in DNS
        rcode, msg, soa_serial, *stuff = update_session.update_zone(
                            self.zone_name, zi, 
                            force_soa_serial_update=not(settings['no_serial']),
                            wrap_serial_next_time=settings['wrap_serial'],
                            nsec3_seed=settings['nsec3_seed'],
                            clear_nsec3=settings['clear_nsec3'],
                            clear_dnskey=settings['clear_dnskey']
                            )

        if rcode == RCODE_NOCHANGE:
            log_info(msg)
            sys.exit(os.EX_OK)

        # Delete temporary file
        clean_up()

        if rcode == RCODE_ERROR:
            log_warning(msg)
            sys.exit(os.EX_TEMPFAIL)
        elif rcode == RCODE_RESET:
            log_error(msg)
            sys.exit(os.EX_IOERR)
        elif rcode == RCODE_FATAL:
            log_error(msg)
            sys.exit(os.EX_IOERR)

        # Everything good - Lets GO!
        if (settings['verbose']):
            log_info(msg)
        else:
            log_debug(msg)
        sys.exit(os.EX_OK)




if (__name__ is "__main__"):
    exit_code = DynDNSTool(sys.argv, len(sys.argv))
    sys.exit(exit_code)




#!/usr/bin/env python3.2
#
# Copyright (c) Net24 Limited, Christchurch, New Zealand 2011-2012
#       and     Voyager Internet Ltd, New Zealand, 2012-2013
#
#    This file is part of py-magcode-core.
#
#    Py-magcode-core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU  General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Py-magcode-core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU  General Public License for more details.
#
#    You should have received a copy of the GNU  General Public License
#    along with py-magcode-core.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Base Update Engine class.  Contains common code.
"""

import socket
from datetime import datetime

import dns.query
import dns.resolver
import dns.zone
import dns.exception
import dns.rdatatype
import dns.name
import dns.message
import dns.rdataclass
import dns.flags

from dms.exceptions import SOASerialArithmeticError
from dms.exceptions import NoSuchZoneOnServerError
from magcode.core.globals_ import settings
from magcode.core.database import RCODE_FATAL
from magcode.core.database import RCODE_RESET
from magcode.core.database import RCODE_OK
from magcode.core.database import RCODE_ERROR
from dms.dns import RRTYPE_SOA
from dms.dns import RRTYPE_NSEC3PARAM
from dms.dns import RRTYPE_DNSKEY
# initialise settings keys
import dms.globals_


class UpdateEngine(object):
    """
    Parent Generic Update Engine class

    Contains common code etc.
    """
    def __init__(self, server, port=None):
        """
        Initialise settings for conversation.
        """
        # Make sure we can connect to server port TCP 53
        # NOTE: can be IPv4 or V6 - depends on resolv.conf option inet6 setting
        dest_port = settings['dns_port']
        if (port):            
            dest_port = port
        (family, type_, proto, canonname, sockaddr) \
            = socket.getaddrinfo(server, dest_port, proto=socket.SOL_TCP)[0]
        sock = socket.socket(family=family, type=type_, proto=proto)
        sock.connect(sockaddr)
        sock.close()
        # If we get here without raising an exception, we can talk to the 
        # server (mostly)
        self.server_name = server
        self.port_name = dest_port
        self.server = sockaddr[0]
        self.port = sockaddr[1]

    def read_zone(self, zone_name, filter_dnssec=True):
        """
        Use dnspython to read in a Zone from the DNS server
        
        Returns a Zone Instance based on the read in data.
        NOTE: This is not from the DB!
        """
        xfr_generator = dns.query.xfr(self.server, zone_name, port=self.port)
        try:
            zone = dns.zone.from_xfr(xfr_generator)
        except dns.exception.FormError as exc:
            zone = None
        finally:
            del xfr_generator
        # Unlink exception chaining
        if (not zone):
            raise NoSuchZoneOnServerError(zone_name, self.server_name, 
                                            self.port)

        # Filter out dnssec if requested.
        dnssec_types = settings['dnssec_filter'].split()
        dnssec_rdtypes = [dns.rdatatype.from_text(x) for x in dnssec_types]
        nsec3param_rdtype = dns.rdatatype.from_text(RRTYPE_NSEC3PARAM)
        dnskey_rdtype = dns.rdatatype.from_text(RRTYPE_DNSKEY)
        # Need to find items to delete before deleting them, or
        # else zone data structure is corrupted.
        rr_delete_list = []
        dnskey_flag = False
        nsec3param_flag = False
        for rdata in zone.iterate_rdatas():
            if not dnskey_flag:
                dnskey_flag = (rdata[2].rdtype == dnskey_rdtype)
            if not nsec3param_flag:
                nsec3param_flag = (rdata[2].rdtype == nsec3param_rdtype)
            if rdata[2].rdtype in dnssec_rdtypes:
                rr_delete_list.append((rdata[0], rdata[2].rdtype, 
                        rdata[2].covers(),))
        # Finally delete all unwanted records
        if filter_dnssec:
            for (name, rdtype, covers) in rr_delete_list:
                zone.delete_rdataset(name, rdtype, covers)
        # Finally, an unclutered zone without DNSSEC
        return (zone, dnskey_flag, nsec3param_flag)

    def read_soa(self, zone_name):
        """
        Use dnspython to read the SOA record of a Zone from the DNS server.

        Returns the SOA serial number etc.  This is intended as a test
        function to see if the master server has configured a zone.
        """
        zone = dns.name.from_text(zone_name)
        rdtype = dns.rdatatype.from_text(RRTYPE_SOA)
        rdclass = dns.rdataclass.IN
        query = dns.message.make_query(zone, rdtype, rdclass)
        exc = None
        try:
            # Use TCP as dnspython can't track multi-threaded udp query/results
            answer = dns.query.tcp(query, self.server, port=self.port,
                    timeout=float(settings['dns_query_timeout']))
        except dns.exception.Timeout:
            msg =  ("Zone '%s', - timeout waiting for response, retrying"
                    % zone_name)
            return (RCODE_ERROR, msg, None)
        except (dns.query.UnexpectedSource, dns.query.BadResponse,
                dns.query.FormError) as exc:
            # For UDP, FormError and BadResponse here are retrys as they
            # mostly could be transitory
            msg = ("Zone '%s', - reply from unexpected source, retrying"
                    % zone_name)
            return (RCODE_ERROR, msg, None)
        # Here to show what should be done if any of above errors prove
        # to be more than a retry...
        #except (dns.query.BadResponse,
        #        dns.exception.FormError) as exc:
        #    msg = ("Zone '%s', - server %s not operating correctly." 
        #                % (zone_name, server.server_name))
        #    return (RCODE_FATAL, msg, None)
        except socket.error as exc:
            if errno in (errno.EACCESS, errno.EPERM, errno.ECONNREFUSED, 
                    errno.ENETUNREACHABLE, errno.ETIMEDOUT):
                msg = ("Zone '%s' - can't reach server %s:%s yet - %s"
                        % (zone_name, self.server, self.port, exc.strerror))
                return (RCODE_ERROR, msg, None)
            msg = ("Zone '%s' - server %s:%s, fatal error %s."
                   % (zone_name, self.server, self.port, exc.strerror))
            return (RCODE_FATAL, msg, None)
        finally:
            # Clean up memory
            del query
        try:
            if (answer.flags & dns.flags.AA != dns.flags.AA):
                msg = "Zone '%s' not yet operational on server." % zone_name
                return (RCODE_RESET, msg, None)
            if (len(answer.answer) != 1):
                msg = "Zone '%s' not yet operational on server." % zone_name
                return (RCODE_RESET, msg, None)
            if not len(answer.answer[0].items):
                msg = "Zone '%s' not yet operational on server." % zone_name
                return (RCODE_RESET, msg, None)
            if answer.answer[0].items[0].rdtype != rdtype:
                msg = "Zone '%s' not yet operational on server." % zone_name
                return (RCODE_RESET, msg, None)
            
            # We succeeded in getting an SOA and serial number
            msg = "Zone '%s' operational on server." % zone_name 
            return (RCODE_OK, msg, answer.answer[0].items[0].serial)
        finally:
            # clean up memory
            del answer

    def get_serial_no(self, zone):
        """
        Obtain the serial number from the SOA of a zone
        """
        rdataset = zone.find_rdataset(zone.origin, 
                                        dns.rdatatype.from_text(RRTYPE_SOA))
        return rdataset.items[0].serial
       
    def update_zone(self, zone_name, zi, db_soa_serial=None, 
            candidate_soa_serial=None,
            force_soa_serial_update=False, wrap_serial_next_time=False,
            date_stamp=None, nsec3_seed=False, clear_dnskey=False,
            clear_nsec3=False):
        """
        Stub method for updating a zone.

        returns tuple of RCODE_, message, serial number
        """
        message = 'Stub update_zone()'
        return (RCODE_FATAL, message, None, None)
